#!/usr/bin/env python3

import os
import re
import subprocess
import sys
import argparse
from pathlib import Path
from typing import Tuple, Optional
import tomllib

class SemverPublisher:
    def __init__(self, project_root: str = "."):
        self.project_root = Path(project_root).absolute()
        self.pyproject_path = self.project_root / "pyproject.toml"

    def get_current_version(self) -> str:
        if not self.pyproject_path.exists():
            raise FileNotFoundError("pyproject.toml not found")

        with open(self.pyproject_path, "rb") as f:
            data = tomllib.load(f)

        return data["project"]["version"]

    def parse_semver(self, version: str) -> Tuple[int, int, int]:
        pattern = r"^(\d+)\.(\d+)\.(\d+)(?:-.*)?(?:\+.*)?$"
        match = re.match(pattern, version)
        if not match:
            raise ValueError(f"Invalid semver format: {version}")
        return int(match.group(1)), int(match.group(2)), int(match.group(3))

    def increment_version(self, version: str, bump_type: str) -> str:
        major, minor, patch = self.parse_semver(version)

        if bump_type == "major":
            return f"{major + 1}.0.0"
        elif bump_type == "minor":
            return f"{major}.{minor + 1}.0"
        elif bump_type == "patch":
            return f"{major}.{minor}.{patch + 1}"
        else:
            raise ValueError(f"Invalid bump type: {bump_type}")

    def update_pyproject_version(self, new_version: str) -> None:
        with open(self.pyproject_path, "r") as f:
            content = f.read()

        pattern = r'(\[project\][\s\S]*?)version\s*=\s*"[^"]+"'
        def replacement_func(match):
            project_section = match.group(1)
            return f'{project_section}version = "{new_version}"'

        new_content = re.sub(pattern, replacement_func, content)

        if new_content == content:
            raise ValueError("Version pattern not found in pyproject.toml")

        if 'target-version = "' + new_version + '"' in new_content:
            raise ValueError(f"Version update would corrupt tool configuration. Ruff target-version should not be changed to package version {new_version}")

        with open(self.pyproject_path, "w") as f:
            f.write(new_content)

    def run_command(self, cmd: list, cwd: Optional[Path] = None) -> subprocess.CompletedProcess:
        try:
            result = subprocess.run(
                cmd,
                cwd=cwd or self.project_root,
                capture_output=True,
                text=True,
                check=True
            )
            return result
        except subprocess.CalledProcessError as e:
            print(f"Command failed: {' '.join(cmd)}")
            print(f"Error: {e.stderr}")
            raise

    def clean_build(self) -> None:
        build_dirs = [self.project_root / "build", self.project_root / "dist"]
        for build_dir in build_dirs:
            if build_dir.exists():
                self.run_command(["rm", "-rf", str(build_dir)])

    def build_package(self) -> None:
        self.run_command([sys.executable, "-m", "build"])

    def validate_package(self) -> None:
        self.run_command(["twine", "check", "dist/*"])

    def run_tests(self) -> None:
        self.run_command([sys.executable, "-m", "unittest", "discover", "-s", "tests"])

    def publish_to_pypi(self, repository: str = "pypi") -> None:
        cmd = ["twine", "upload"]
        if repository != "pypi":
            cmd.extend(["--repository", repository])
        cmd.append("dist/*")

        self.run_command(cmd)

    def git_commit_and_tag(self, version: str, message: Optional[str] = None) -> None:
        if not message:
            message = f"Release v{version}"

        self.run_command(["git", "add", "pyproject.toml"])
        self.run_command(["git", "commit", "-m", message])
        self.run_command(["git", "tag", f"v{version}"])

    def check_git_status(self) -> bool:
        try:
            result = self.run_command(["git", "status", "--porcelain"])
            return len(result.stdout.strip()) == 0
        except subprocess.CalledProcessError:
            return False

    def publish(self, bump_type: str, dry_run: bool = False, skip_tests: bool = False,
                repository: str = "pypi", commit_message: Optional[str] = None) -> str:

        print("🚀 Starting semver-based package publishing...")

        if not self.check_git_status() and not dry_run:
            print("⚠️  Warning: Git working directory is not clean")
            response = input("Continue anyway? (y/N): ")
            if response.lower() != 'y':
                print("Aborted.")
                return ""

        current_version = self.get_current_version()
        new_version = self.increment_version(current_version, bump_type)

        print(f"📦 Current version: {current_version}")
        print(f"📦 New version: {new_version}")

        if dry_run:
            print("🔍 DRY RUN - No changes will be made")
            return new_version

        try:
            if not skip_tests:
                print("🧪 Running tests...")
                self.run_tests()
                print("✅ Tests passed")

            print(f"📝 Updating version to {new_version}...")
            self.update_pyproject_version(new_version)

            print("🧹 Cleaning build directories...")
            self.clean_build()

            print("🔨 Building package...")
            self.build_package()

            print("🔍 Validating package...")
            self.validate_package()

            print(f"📤 Publishing to {repository}...")
            self.publish_to_pypi(repository)

            print("📝 Committing and tagging...")
            self.git_commit_and_tag(new_version, commit_message)

            print(f"🎉 Successfully published {new_version}!")
            print(f"🏷️  Git tag: v{new_version}")

            if repository == "pypi":
                print(f"🌐 PyPI: https://pypi.org/project/mcp-mathematics/{new_version}/")

            return new_version

        except Exception as e:
            print(f"❌ Publishing failed: {e}")
            print("💡 You may need to manually revert version changes")
            raise

def main():
    parser = argparse.ArgumentParser(description="Semver-based package publishing")
    parser.add_argument("bump_type", choices=["major", "minor", "patch"],
                       help="Version bump type")
    parser.add_argument("--dry-run", action="store_true",
                       help="Show what would be done without making changes")
    parser.add_argument("--skip-tests", action="store_true",
                       help="Skip running tests")
    parser.add_argument("--repository", default="pypi",
                       help="PyPI repository (default: pypi)")
    parser.add_argument("--message",
                       help="Custom commit message")
    parser.add_argument("--project-root", default=".",
                       help="Project root directory")

    args = parser.parse_args()

    try:
        publisher = SemverPublisher(args.project_root)
        new_version = publisher.publish(
            bump_type=args.bump_type,
            dry_run=args.dry_run,
            skip_tests=args.skip_tests,
            repository=args.repository,
            commit_message=args.message
        )

        if not args.dry_run:
            print(f"\n✨ Next steps:")
            print(f"   git push origin main")
            print(f"   git push origin v{new_version}")

    except KeyboardInterrupt:
        print("\n❌ Aborted by user")
        sys.exit(1)
    except Exception as e:
        print(f"\n❌ Error: {e}")
        sys.exit(1)

if __name__ == "__main__":
    main()


import ast
import cmath
import datetime
import logging
import math
import operator
import queue
import re
import resource
import statistics
import threading
import time
from collections import OrderedDict, deque
from contextlib import contextmanager
from dataclasses import dataclass, field
from threading import RLock, Timer
from typing import Any

from mcp.server import FastMCP

MAXIMUM_MATHEMATICAL_EXPRESSION_CHARACTER_LIMIT = 1000
MAXIMUM_AST_NODE_DEPTH_LIMIT = 10
CALCULATION_HISTORY_ENTRY_LIMIT = 100
FACTORIAL_COMPUTATION_UPPER_BOUND = 300
EXPONENTIATION_SAFETY_THRESHOLD = 10000
MAXIMUM_COMPUTATION_RESULT_CHARACTER_LIMIT = 10000
MAXIMUM_MEMORY_USAGE_MEGABYTES = 512
MAXIMUM_LIST_ELEMENT_COUNT_LIMIT = 10000

MATHEMATICAL_OPERATION_TIMEOUT_SECONDS = 15.0
SECURITY_AUDIT_LOGGING_ENABLED = True

MAXIMUM_COMPUTATION_DURATION_SECONDS = 20.0
MAXIMUM_RECURSIVE_FUNCTION_CALL_DEPTH = 1000
ENABLE_RATE_LIMITING = True
RATE_LIMIT_WINDOW = 60
MAXIMUM_CLIENT_REQUESTS_PER_TIME_WINDOW = 1000000
ENABLE_INPUT_HASHING = True

FORBIDDEN_PATTERNS = [
    r"import\s+",
    r"exec\s*\(",
    r"eval\s*\(",
    r"__.*__",
    r"globals\s*\(",
    r"locals\s*\(",
    r"getattr\s*\(",
    r"setattr\s*\(",
    r"delattr\s*\(",
    r"hasattr\s*\(",
]

_expression_cache_lock = RLock()
_computation_metrics_lock = RLock()
_calculation_history_lock = RLock()
_session_variables_lock = RLock()
_rate_limiting_lock = RLock()

CACHE_TTL = 300
MAX_CACHE_SIZE = 1000
MAX_SESSION_TTL = 3600
MAX_SESSIONS = 100
MAX_COMPUTATION_STATS = 500
SESSION_CLEANUP_INTERVAL = 600

_memory_cleanup_timer = None

mathematical_computation_logger = logging.getLogger(__name__)
mathematical_security_logger = logging.getLogger(f"{__name__}.security")
mathematical_security_logger.setLevel(logging.INFO)


class AsynchronousLoggingManager:
    def __init__(self, base_logger):
        self.queue = queue.Queue()
        self.logger = base_logger
        self.worker = threading.Thread(target=self._process_logs, daemon=True)
        self.worker.start()

    def log(self, level, message):
        self.queue.put((level, message, time.time()))

    def _process_logs(self):
        while True:
            try:
                level, message, timestamp = self.queue.get(timeout=0.1)
                self.logger.log(level, f"[{timestamp:.3f}] {message}")
            except queue.Empty:
                continue


async_mathematical_logger = AsynchronousLoggingManager(mathematical_computation_logger)
async_mathematical_security_logger = AsynchronousLoggingManager(mathematical_security_logger)


class LeastRecentlyUsedCache:
    def __init__(self, max_size: int = MAX_CACHE_SIZE):
        self.max_size = max_size
        self.cache = OrderedDict()
        self.lock = RLock()

    def get(self, key: str) -> Any:
        with self.lock:
            if key not in self.cache:
                return None
            self.cache.move_to_end(key)
            return self.cache[key]

    def set(self, key: str, value: Any) -> None:
        with self.lock:
            if key in self.cache:
                self.cache.move_to_end(key)
            elif len(self.cache) >= self.max_size:
                self.cache.popitem(last=False)
            self.cache[key] = value

    def clear(self) -> None:
        with self.lock:
            self.cache.clear()

    def size(self) -> int:
        with self.lock:
            return len(self.cache)


class TimeToLiveExpirationCache:
    def __init__(self, max_size: int = MAX_CACHE_SIZE, ttl: float = CACHE_TTL):
        self.max_size = max_size
        self.ttl = ttl
        self.cache = OrderedDict()
        self.lock = RLock()

    def get(self, key: str) -> Any:
        with self.lock:
            if key not in self.cache:
                return None
            value, timestamp = self.cache[key]
            if time.time() - timestamp > self.ttl:
                del self.cache[key]
                return None
            self.cache.move_to_end(key)
            return value

    def set(self, key: str, value: Any) -> None:
        with self.lock:
            current_time = time.time()
            if key in self.cache:
                self.cache.move_to_end(key)
            elif len(self.cache) >= self.max_size:
                self.cache.popitem(last=False)
            self.cache[key] = (value, current_time)

    def cleanup_expired(self) -> int:
        with self.lock:
            current_time = time.time()
            expired_keys = [
                key
                for key, (_, timestamp) in self.cache.items()
                if current_time - timestamp > self.ttl
            ]
            for key in expired_keys:
                del self.cache[key]
            return len(expired_keys)

    def clear(self) -> None:
        with self.lock:
            self.cache.clear()

    def size(self) -> int:
        with self.lock:
            return len(self.cache)


class UserSessionStateManager:
    def __init__(self, max_sessions: int = MAX_SESSIONS, session_ttl: float = MAX_SESSION_TTL):
        self.max_sessions = max_sessions
        self.session_ttl = session_ttl
        self.sessions = OrderedDict()
        self.session_timestamps = {}
        self.lock = RLock()
        self.cleanup_timer = None
        self._start_cleanup_timer()

    def create_session(self, session_id: str, variables: dict[str, float | complex] = None) -> None:
        with self.lock:
            current_time = time.time()
            if len(self.sessions) >= self.max_sessions:
                oldest_session = next(iter(self.sessions))
                del self.sessions[oldest_session]
                del self.session_timestamps[oldest_session]

            self.sessions[session_id] = variables or {}
            self.session_timestamps[session_id] = current_time

    def get_session(self, session_id: str) -> dict[str, float | complex] | None:
        with self.lock:
            if session_id not in self.sessions:
                return None
            current_time = time.time()
            if current_time - self.session_timestamps[session_id] > self.session_ttl:
                del self.sessions[session_id]
                del self.session_timestamps[session_id]
                return None
            self.session_timestamps[session_id] = current_time
            self.sessions.move_to_end(session_id)
            return self.sessions[session_id]

    def update_session(self, session_id: str, key: str, value: float | complex) -> bool:
        with self.lock:
            session = self.get_session(session_id)
            if session is None:
                return False
            session[key] = value
            return True

    def delete_session(self, session_id: str) -> bool:
        with self.lock:
            if session_id in self.sessions:
                del self.sessions[session_id]
                del self.session_timestamps[session_id]
                return True
            return False

    def cleanup_expired(self) -> int:
        with self.lock:
            current_time = time.time()
            expired_sessions = [
                session_id
                for session_id, timestamp in self.session_timestamps.items()
                if current_time - timestamp > self.session_ttl
            ]
            for session_id in expired_sessions:
                del self.sessions[session_id]
                del self.session_timestamps[session_id]
            return len(expired_sessions)

    def _start_cleanup_timer(self) -> None:
        def cleanup_task():
            try:
                self.cleanup_expired()
            finally:
                self._start_cleanup_timer()

        self.cleanup_timer = Timer(SESSION_CLEANUP_INTERVAL, cleanup_task)
        self.cleanup_timer.daemon = True
        self.cleanup_timer.start()

    def shutdown(self) -> None:
        if self.cleanup_timer:
            self.cleanup_timer.cancel()

    def get_stats(self) -> dict[str, Any]:
        with self.lock:
            return {
                "active_sessions": len(self.sessions),
                "max_sessions": self.max_sessions,
                "session_ttl": self.session_ttl,
            }


_expression_cache = TimeToLiveExpirationCache(MAX_CACHE_SIZE, CACHE_TTL)
_parsed_expression_ast_cache = LeastRecentlyUsedCache(MAX_CACHE_SIZE)
_computation_stats = LeastRecentlyUsedCache(MAX_COMPUTATION_STATS)
_session_manager = UserSessionStateManager()


@dataclass
class MathematicalComputationResult:
    success: bool
    expression: str
    result: float | complex | bool | tuple | list | None = None
    error: dict[str, Any] | None = None
    metadata: dict[str, Any] = field(default_factory=dict)

    def to_dict(self):
        return {
            "success": self.success,
            "expression": self.expression,
            "result": str(self.result) if self.result is not None else None,
            "error": self.error,
            "metadata": self.metadata,
        }

    def to_string(self):
        if self.success:
            return f"Result: {self.result}"
        else:
            return f"Error: {self.error.get('message', 'Unknown error')}"


class ClientRequestRateLimiter:
    def __init__(self, max_requests: int, window: int, max_clients: int = 1000):
        self.max_requests = max_requests
        self.window = window
        self.max_clients = max_clients
        self.requests = OrderedDict()
        self.lock = RLock()

    def check_rate_limit(self, client_id: str) -> bool:
        current_time = time.time()
        with self.lock:
            if client_id not in self.requests:
                if len(self.requests) >= self.max_clients:
                    oldest_client = next(iter(self.requests))
                    del self.requests[oldest_client]
                self.requests[client_id] = deque()

            client_requests = self.requests[client_id]
            while client_requests and client_requests[0] < current_time - self.window:
                client_requests.popleft()

            if len(client_requests) >= self.max_requests:
                return False

            client_requests.append(current_time)
            self.requests.move_to_end(client_id)
            return True

    def cleanup_expired(self) -> int:
        current_time = time.time()
        with self.lock:
            expired_clients = []
            for client_id, client_requests in self.requests.items():
                while client_requests and client_requests[0] < current_time - self.window:
                    client_requests.popleft()
                if not client_requests:
                    expired_clients.append(client_id)

            for client_id in expired_clients:
                del self.requests[client_id]
            return len(expired_clients)


mathematical_computation_rate_limiter = ClientRequestRateLimiter(MAXIMUM_CLIENT_REQUESTS_PER_TIME_WINDOW, RATE_LIMIT_WINDOW)


@contextmanager
def mathematical_computation_timeout(seconds: float):
    timeout_occurred = threading.Event()

    def timeout_callback():
        timeout_occurred.set()

    timer = threading.Timer(seconds, timeout_callback)
    timer.start()

    try:
        yield timeout_occurred
    finally:
        timer.cancel()
        if timeout_occurred.is_set():
            raise TimeoutError(f"Computation exceeded {seconds}s limit")


@contextmanager
def mathematical_computation_memory_limit(mb: int):
    try:
        soft, hard = resource.getrlimit(resource.RLIMIT_AS)
        new_limit = mb * 1024 * 1024

        if new_limit < soft and new_limit < hard:
            resource.setrlimit(resource.RLIMIT_AS, (new_limit, hard))
            try:
                yield
            finally:
                resource.setrlimit(resource.RLIMIT_AS, (soft, hard))
        else:
            yield
    except (ValueError, OSError):
        yield


OPERATORS = {
    ast.Add: operator.add,
    ast.Sub: operator.sub,
    ast.Mult: operator.mul,
    ast.Div: operator.truediv,
    ast.FloorDiv: operator.floordiv,
    ast.Mod: operator.mod,
    ast.Pow: operator.pow,
    ast.USub: operator.neg,
    ast.UAdd: operator.pos,
    ast.BitXor: operator.xor,
    ast.BitOr: operator.or_,
    ast.BitAnd: operator.and_,
    ast.LShift: operator.lshift,
    ast.RShift: operator.rshift,
}

COMPARISON_OPS = {
    ast.Eq: operator.eq,
    ast.NotEq: operator.ne,
    ast.Lt: operator.lt,
    ast.LtE: operator.le,
    ast.Gt: operator.gt,
    ast.GtE: operator.ge,
}

MATH_FUNCTIONS = {
    "sin": math.sin,
    "cos": math.cos,
    "tan": math.tan,
    "asin": math.asin,
    "acos": math.acos,
    "atan": math.atan,
    "atan2": math.atan2,
    "sinh": math.sinh,
    "cosh": math.cosh,
    "tanh": math.tanh,
    "asinh": math.asinh,
    "acosh": math.acosh,
    "atanh": math.atanh,
    "log": math.log,
    "log10": math.log10,
    "log2": math.log2,
    "exp": math.exp,
    "sqrt": math.sqrt,
    "pow": math.pow,
    "abs": lambda x: abs(x),
    "fabs": math.fabs,
    "factorial": math.factorial,
    "ceil": math.ceil,
    "floor": math.floor,
    "trunc": math.trunc,
    "degrees": math.degrees,
    "radians": math.radians,
    "gcd": math.gcd,
    "lcm": getattr(math, "lcm", None),
    "isqrt": getattr(math, "isqrt", None),
    "hypot": math.hypot,
    "copysign": math.copysign,
    "fmod": math.fmod,
    "remainder": math.remainder,
    "modf": math.modf,
    "frexp": math.frexp,
    "ldexp": math.ldexp,
    "isfinite": math.isfinite,
    "isinf": math.isinf,
    "isnan": math.isnan,
    "isclose": math.isclose,
    "comb": getattr(math, "comb", None),
    "perm": getattr(math, "perm", None),
    "erf": math.erf,
    "erfc": math.erfc,
    "gamma": math.gamma,
    "lgamma": math.lgamma,
    "cbrt": getattr(math, "cbrt", None),
    "exp2": getattr(math, "exp2", None),
    "expm1": math.expm1,
    "log1p": math.log1p,
    "nextafter": getattr(math, "nextafter", None),
    "ulp": getattr(math, "ulp", None),
}

STATISTICS_FUNCTIONS = {
    "mean": statistics.mean,
    "median": statistics.median,
    "mode": statistics.mode,
    "stdev": statistics.stdev,
    "pstdev": statistics.pstdev,
    "variance": statistics.variance,
    "pvariance": statistics.pvariance,
    "harmonic_mean": statistics.harmonic_mean,
    "geometric_mean": getattr(statistics, "geometric_mean", None),
    "quantiles": getattr(statistics, "quantiles", None),
}

COMPLEX_FUNCTIONS = {
    "phase": cmath.phase,
    "polar": cmath.polar,
    "rect": cmath.rect,
    "csin": cmath.sin,
    "ccos": cmath.cos,
    "ctan": cmath.tan,
    "cexp": cmath.exp,
    "clog": cmath.log,
    "clog10": cmath.log10,
    "csqrt": cmath.sqrt,
}

UNIT_CONVERSIONS = {
    "length": {
        "m": 1.0,
        "km": 1000.0,
        "cm": 0.01,
        "mm": 0.001,
        "mi": 1609.344,
        "yd": 0.9144,
        "ft": 0.3048,
        "in": 0.0254,
        "nmi": 1852.0,
        "ly": 9.4607e15,
        "AU": 1.496e11,
        "pc": 3.0857e16,
        "angstrom": 1e-10,
        "micron": 1e-6,
        "nm": 1e-9,
    },
    "mass": {
        "kg": 1.0,
        "g": 0.001,
        "mg": 1e-6,
        "lb": 0.453592,
        "oz": 0.0283495,
        "ton": 1000.0,
        "t": 1000.0,
        "ton_us": 907.185,
        "ton_uk": 1016.05,
        "st": 6.35029,
        "ct": 0.0002,
        "gr": 6.47989e-5,
        "amu": 1.66054e-27,
    },
    "time": {
        "s": 1.0,
        "ms": 0.001,
        "us": 1e-6,
        "ns": 1e-9,
        "min": 60.0,
        "h": 3600.0,
        "d": 86400.0,
        "wk": 604800.0,
        "mo": 2592000.0,
        "yr": 31536000.0,
        "decade": 315360000.0,
        "century": 3153600000.0,
        "ps": 1e-12,
        "fortnight": 1209600.0,
        "millennium": 31536000000.0,
    },
    "temperature": {
        "K": lambda x: x,
        "C": lambda x: x + 273.15,
        "F": lambda x: (x - 32) * 5 / 9 + 273.15,
    },
    "area": {
        "m2": 1.0,
        "km2": 1e6,
        "cm2": 1e-4,
        "mm2": 1e-6,
        "ft2": 0.092903,
        "yd2": 0.836127,
        "in2": 0.00064516,
        "mi2": 2.59e6,
        "acre": 4046.86,
        "hectare": 10000.0,
        "are": 100.0,
        "sqch": 404.686,
    },
    "volume": {
        "L": 0.001,
        "mL": 1e-6,
        "m3": 1.0,
        "cm3": 1e-6,
        "gal": 0.00378541,
        "qt": 0.000946353,
        "pt": 0.000473176,
        "fl_oz": 2.95735e-5,
        "gal_uk": 0.00454609,
        "qt_uk": 0.00113652,
        "pt_uk": 0.000568261,
        "cup": 0.000236588,
        "tbsp": 1.47868e-5,
        "tsp": 4.92892e-6,
        "ft3": 0.0283168,
        "in3": 1.63871e-5,
    },
    "speed": {
        "m/s": 1.0,
        "km/h": 0.277778,
        "mph": 0.44704,
        "ft/s": 0.3048,
        "knot": 0.514444,
        "mach": 340.29,
        "cm/s": 0.01,
        "mi/min": 26.8224,
        "in/s": 0.0254,
        "c": 299792458.0,
    },
    "data": {
        "B": 1.0,
        "KB": 1000.0,
        "MB": 1e6,
        "GB": 1e9,
        "TB": 1e12,
        "PB": 1e15,
        "EB": 1e18,
        "ZB": 1e21,
        "bit": 0.125,
        "Kbit": 125.0,
        "Mbit": 125000.0,
        "Gbit": 1.25e8,
        "Tbit": 1.25e11,
        "KiB": 1024.0,
        "MiB": 1048576.0,
        "GiB": 1073741824.0,
    },
    "pressure": {
        "Pa": 1.0,
        "kPa": 1000.0,
        "MPa": 1e6,
        "atm": 101325.0,
        "bar": 100000.0,
        "mbar": 100.0,
        "psi": 6894.76,
        "torr": 133.322,
        "mmHg": 133.322,
        "inHg": 3386.39,
    },
    "energy": {
        "J": 1.0,
        "kJ": 1000.0,
        "MJ": 1e6,
        "cal": 4.184,
        "kcal": 4184.0,
        "Wh": 3600.0,
        "kWh": 3.6e6,
        "BTU": 1055.06,
        "eV": 1.60218e-19,
        "ft_lb": 1.35582,
        "erg": 1e-7,
        "therm": 1.055e8,
    },
    "power": {
        "W": 1.0,
        "kW": 1000.0,
        "MW": 1e6,
        "hp": 745.7,
        "PS": 735.499,
        "BTU/h": 0.293071,
        "ft_lb/s": 1.35582,
        "cal/s": 4.184,
        "erg/s": 1e-7,
        "ton_refrigeration": 3516.85,
    },
    "force": {
        "N": 1.0,
        "kN": 1000.0,
        "lbf": 4.44822,
        "kgf": 9.80665,
        "dyne": 1e-5,
        "pdl": 0.138255,
        "ozf": 0.278014,
        "tonf": 9806.65,
    },
    "angle": {
        "deg": 1.0,
        "rad": 57.2958,
        "grad": 0.9,
        "arcmin": 0.0166667,
        "arcsec": 0.000277778,
        "turn": 360.0,
    },
    "frequency": {
        "Hz": 1.0,
        "kHz": 1000.0,
        "MHz": 1e6,
        "GHz": 1e9,
        "rpm": 0.0166667,
        "rad/s": 0.159155,
    },
    "fuel_economy": {
        "mpg": 1.0,
        "mpg_uk": 1.20095,
        "L/100km": 235.215,
        "km/L": 2.35215,
        "mi/L": 3.78541,
        "gal/100mi": 100.0,
    },
}

UNIT_ALIASES = {
    "meters": "m",
    "meter": "m",
    "metres": "m",
    "metre": "m",
    "kilometers": "km",
    "kilometer": "km",
    "kilometres": "km",
    "kilometre": "km",
    "centimeters": "cm",
    "centimeter": "cm",
    "centimetres": "cm",
    "centimetre": "cm",
    "millimeters": "mm",
    "millimeter": "mm",
    "millimetres": "mm",
    "millimetre": "mm",
    "miles": "mi",
    "mile": "mi",
    "yards": "yd",
    "yard": "yd",
    "feet": "ft",
    "foot": "ft",
    "inches": "in",
    "inch": "in",
    "kilograms": "kg",
    "kilogram": "kg",
    "grams": "g",
    "gram": "g",
    "milligrams": "mg",
    "milligram": "mg",
    "pounds": "lb",
    "pound": "lb",
    "lbs": "lb",
    "ounces": "oz",
    "ounce": "oz",
    "tons": "ton",
    "seconds": "s",
    "second": "s",
    "sec": "s",
    "secs": "s",
    "milliseconds": "ms",
    "millisecond": "ms",
    "millisec": "ms",
    "millisecs": "ms",
    "microseconds": "us",
    "microsecond": "us",
    "microsec": "us",
    "microsecs": "us",
    "nanoseconds": "ns",
    "nanosecond": "ns",
    "nanosec": "ns",
    "nanosecs": "ns",
    "minutes": "min",
    "minute": "min",
    "mins": "min",
    "hours": "h",
    "hour": "h",
    "hrs": "h",
    "hr": "h",
    "days": "d",
    "day": "d",
    "weeks": "wk",
    "week": "wk",
    "months": "mo",
    "month": "mo",
    "years": "yr",
    "year": "yr",
    "yrs": "yr",
    "decades": "decade",
    "centuries": "century",
    "celsius": "C",
    "centigrade": "C",
    "fahrenheit": "F",
    "fahr": "F",
    "kelvin": "K",
    "liters": "L",
    "liter": "L",
    "litres": "L",
    "litre": "L",
    "milliliters": "mL",
    "milliliter": "mL",
    "millilitres": "mL",
    "millilitre": "mL",
    "gallons": "gal",
    "gallon": "gal",
    "quarts": "qt",
    "quart": "qt",
    "pints": "pt",
    "pint": "pt",
    "cups": "cup",
    "tablespoons": "tbsp",
    "tablespoon": "tbsp",
    "teaspoons": "tsp",
    "teaspoon": "tsp",
    "bytes": "B",
    "byte": "B",
    "kilobytes": "KB",
    "kilobyte": "KB",
    "megabytes": "MB",
    "megabyte": "MB",
    "gigabytes": "GB",
    "gigabyte": "GB",
    "terabytes": "TB",
    "terabyte": "TB",
    "petabytes": "PB",
    "petabyte": "PB",
    "bits": "bit",
    "kilobits": "Kbit",
    "kilobit": "Kbit",
    "megabits": "Mbit",
    "megabit": "Mbit",
    "gigabits": "Gbit",
    "gigabit": "Gbit",
    "pascals": "Pa",
    "pascal": "Pa",
    "kilopascals": "kPa",
    "kilopascal": "kPa",
    "megapascals": "MPa",
    "megapascal": "MPa",
    "atmospheres": "atm",
    "atmosphere": "atm",
    "bars": "bar",
    "millibars": "mbar",
    "millibar": "mbar",
    "joules": "J",
    "joule": "J",
    "kilojoules": "kJ",
    "kilojoule": "kJ",
    "megajoules": "MJ",
    "megajoule": "MJ",
    "calories": "cal",
    "calorie": "cal",
    "kilocalories": "kcal",
    "kilocalorie": "kcal",
    "watts": "W",
    "watt": "W",
    "kilowatts": "kW",
    "kilowatt": "kW",
    "megawatts": "MW",
    "megawatt": "MW",
    "horsepower": "hp",
    "newtons": "N",
    "newton": "N",
    "kilonewtons": "kN",
    "kilonewton": "kN",
    "degrees": "deg",
    "degree": "deg",
    "radians": "rad",
    "radian": "rad",
    "hertz": "Hz",
    "kilohertz": "kHz",
    "megahertz": "MHz",
    "gigahertz": "GHz",
}

MATH_CONSTANTS = {
    "pi": math.pi,
    "e": math.e,
    "tau": math.tau,
    "inf": math.inf,
    "nan": math.nan,
    "phi": (1 + math.sqrt(5)) / 2,
    "euler": 0.5772156649,
}

MATH_FUNCTIONS = {k: v for k, v in MATH_FUNCTIONS.items() if v is not None}
STATISTICS_FUNCTIONS = {k: v for k, v in STATISTICS_FUNCTIONS.items() if v is not None}

ALL_FUNCTIONS = {**MATH_FUNCTIONS, **STATISTICS_FUNCTIONS, **COMPLEX_FUNCTIONS}


class MathematicalCalculationHistory:
    def __init__(self, max_size: int = CALCULATION_HISTORY_ENTRY_LIMIT):
        self.history: list[dict[str, Any]] = []
        self.max_size = max_size
        self.lock = RLock()

    def add(self, expression: str, result: str) -> None:
        with self.lock:
            entry = {
                "expression": expression,
                "result": result,
                "timestamp": datetime.datetime.now().isoformat(),
            }
            self.history.append(entry)
            if len(self.history) > self.max_size:
                self.history.pop(0)

            if SECURITY_AUDIT_LOGGING_ENABLED:
                async_mathematical_logger.log(
                    logging.INFO, f"Added calculation: {expression} = {result}"
                )

    def get_recent(self, limit: int = 10) -> list[dict[str, Any]]:
        with self.lock:
            return self.history[-limit:]

    def clear(self) -> None:
        with self.lock:
            if SECURITY_AUDIT_LOGGING_ENABLED:
                async_mathematical_logger.log(logging.INFO, "Calculation history cleared")
            self.history.clear()


mathematical_calculation_history = MathematicalCalculationHistory()


def compute_expression_fingerprint(expression: str) -> str:
    return str(hash(expression))


def validate_expression_security_constraints(expression: str, client_id: str = "default") -> bool:
    if ENABLE_RATE_LIMITING and not mathematical_computation_rate_limiter.check_rate_limit(
        client_id
    ):
        async_mathematical_security_logger.log(
            logging.WARNING, f"Rate limit exceeded for client: {client_id}"
        )
        raise PermissionError(
            f"Rate limit exceeded: {MAXIMUM_CLIENT_REQUESTS_PER_TIME_WINDOW} requests per {RATE_LIMIT_WINDOW} seconds"
        )

    for pattern in FORBIDDEN_PATTERNS:
        if re.search(pattern, expression, re.IGNORECASE):
            async_mathematical_security_logger.log(
                logging.ERROR, f"Forbidden pattern detected: {pattern}"
            )
            raise ValueError("Expression contains forbidden pattern")

    suspicious_chars = ["\\x00", "\\n", "\\r", "\\b", "\\f", "\\v"]
    filtered_expr = "".join(ch for ch in expression if ch not in suspicious_chars)

    if len(filtered_expr) != len(expression):
        async_mathematical_security_logger.log(logging.ERROR, "Control characters detected")
        raise ValueError("Expression contains invalid control characters")

    return True


def track_mathematical_operation_performance(expression: str, start_time: float, end_time: float) -> None:
    with _computation_metrics_lock:
        computation_time = end_time - start_time
        expression_hash = compute_expression_fingerprint(expression)

        existing_stats = _computation_stats.get(expression_hash)
        if existing_stats is None:
            stats = {
                "count": 1,
                "total_time": computation_time,
                "max_time": computation_time,
                "first_seen": start_time,
                "last_seen": end_time,
            }
        else:
            stats = existing_stats.copy()
            stats["count"] += 1
            stats["total_time"] += computation_time
            stats["max_time"] = max(stats["max_time"], computation_time)
            stats["last_seen"] = end_time

        _computation_stats.set(expression_hash, stats)

        if computation_time > MAXIMUM_COMPUTATION_DURATION_SECONDS:
            async_mathematical_security_logger.log(
                logging.WARNING, f"Slow computation: {computation_time:.3f}s for {expression_hash}"
            )


def cleanup_expired_cache_entries() -> dict[str, int]:
    cleanup_stats = {
        "expression_cache_expired": 0,
        "sessions_expired": 0,
        "rate_limiter_expired": 0,
    }

    cleanup_stats["expression_cache_expired"] = _expression_cache.cleanup_expired()
    cleanup_stats["sessions_expired"] = _session_manager.cleanup_expired()
    cleanup_stats["rate_limiter_expired"] = mathematical_computation_rate_limiter.cleanup_expired()

    return cleanup_stats


def start_memory_cleanup_timer() -> None:
    global _memory_cleanup_timer

    def memory_cleanup_task():
        try:
            cleanup_stats = cleanup_expired_cache_entries()
            total_cleaned = sum(cleanup_stats.values())
            if total_cleaned > 0:
                async_mathematical_logger.log(logging.INFO, f"Memory cleanup: {cleanup_stats}")
        except Exception as e:
            async_mathematical_logger.log(logging.ERROR, f"Memory cleanup error: {e}")
        finally:
            start_memory_cleanup_timer()

    _memory_cleanup_timer = Timer(SESSION_CLEANUP_INTERVAL, memory_cleanup_task)
    _memory_cleanup_timer.daemon = True
    _memory_cleanup_timer.start()


def get_memory_usage_stats() -> dict[str, Any]:
    import os

    import psutil

    process = psutil.Process(os.getpid())
    memory_info = process.memory_info()

    return {
        "process_memory_mb": round(memory_info.rss / 1024 / 1024, 2),
        "virtual_memory_mb": round(memory_info.vms / 1024 / 1024, 2),
        "expression_cache_size": _expression_cache.size(),
        "ast_cache_size": _parsed_expression_ast_cache.size(),
        "computation_stats_size": _computation_stats.size(),
        "active_sessions": _session_manager.get_stats()["active_sessions"],
        "rate_limiter_clients": len(mathematical_computation_rate_limiter.requests),
        "history_entries": len(mathematical_calculation_history.history),
    }


def shutdown_memory_management() -> None:
    global _memory_cleanup_timer
    if _memory_cleanup_timer:
        _memory_cleanup_timer.cancel()
    _session_manager.shutdown()


def retrieve_cached_computation_result(expression: str) -> str | None:
    return _expression_cache.get(expression)


def persist_computation_result_in_cache(expression: str, result: str) -> None:
    _expression_cache.set(expression, result)


def retrieve_cached_abstract_syntax_tree(expression: str) -> ast.AST | None:
    return _parsed_expression_ast_cache.get(expression)


def store_parsed_expression(expression: str, tree: ast.AST) -> None:
    _parsed_expression_ast_cache.set(expression, tree)


def sanitize_expression(expr: str) -> str:
    sanitized = "".join(char for char in expr if ord(char) >= 32 or char in " \\t")
    sanitized = " ".join(sanitized.split())

    dangerous_sequences = ["..", "__", "$$", "@@", "##"]
    for seq in dangerous_sequences:
        sanitized = sanitized.replace(seq, "")

    return sanitized


def preprocess_expression(expr: str) -> str:
    expr = expr.replace("×", "*")
    expr = expr.replace("÷", "/")
    expr = expr.replace("^", "**")
    return expr


def validate_ast_node(node: ast.AST, depth: int = 0, session_vars: dict | None = None) -> bool:
    if depth > MAXIMUM_AST_NODE_DEPTH_LIMIT:
        raise ValueError(f"Expression too complex (max depth: {MAXIMUM_AST_NODE_DEPTH_LIMIT})")

    if isinstance(node, ast.Expression):
        return validate_ast_node(node.body, depth + 1, session_vars)

    if isinstance(node, ast.Constant):
        return isinstance(node.value, (int, float, complex, str))

    if isinstance(node, ast.Name):
        if session_vars and node.id in session_vars:
            return True
        return node.id in MATH_CONSTANTS or node.id in ALL_FUNCTIONS

    if isinstance(node, ast.UnaryOp):
        return type(node.op) in OPERATORS and validate_ast_node(
            node.operand, depth + 1, session_vars
        )

    if isinstance(node, ast.BinOp):
        return (
            type(node.op) in OPERATORS
            and validate_ast_node(node.left, depth + 1, session_vars)
            and validate_ast_node(node.right, depth + 1, session_vars)
        )

    if isinstance(node, ast.Compare):
        if len(node.ops) != 1 or len(node.comparators) != 1:
            return False
        return (
            type(node.ops[0]) in COMPARISON_OPS
            and validate_ast_node(node.left, depth + 1, session_vars)
            and validate_ast_node(node.comparators[0], depth + 1, session_vars)
        )

    if isinstance(node, ast.Call):
        if not isinstance(node.func, ast.Name):
            return False
        if node.func.id not in ALL_FUNCTIONS:
            return False
        if node.keywords:
            return False
        return all(validate_ast_node(arg, depth + 1, session_vars) for arg in node.args)

    if isinstance(node, ast.List):
        if len(node.elts) > MAXIMUM_LIST_ELEMENT_COUNT_LIMIT:
            raise ValueError(f"List too large (max size: {MAXIMUM_LIST_ELEMENT_COUNT_LIMIT})")
        return all(validate_ast_node(elt, depth + 1, session_vars) for elt in node.elts)

    return False


def evaluate_mathematical_node(node: ast.AST, session_vars: dict | None = None) -> Any:
    if isinstance(node, ast.Expression):
        return evaluate_mathematical_node(node.body, session_vars)

    if isinstance(node, ast.Constant):
        if isinstance(node.value, str):
            if node.value.endswith("j"):
                return complex(node.value)
            return node.value
        return node.value

    if isinstance(node, ast.Name):
        if session_vars and node.id in session_vars:
            return session_vars[node.id]
        if node.id in MATH_CONSTANTS:
            return MATH_CONSTANTS[node.id]
        raise ValueError(f"Unknown variable: {node.id}")

    if isinstance(node, ast.UnaryOp):
        op = OPERATORS[type(node.op)]
        operand = evaluate_mathematical_node(node.operand, session_vars)
        return op(operand)

    if isinstance(node, ast.BinOp):
        op = OPERATORS[type(node.op)]
        left = evaluate_mathematical_node(node.left, session_vars)
        right = evaluate_mathematical_node(node.right, session_vars)

        if isinstance(node.op, ast.Pow) and abs(right) > EXPONENTIATION_SAFETY_THRESHOLD:
            raise ValueError(f"Power exponent too large (max: {EXPONENTIATION_SAFETY_THRESHOLD})")

        try:
            return op(left, right)
        except (OverflowError, ZeroDivisionError) as e:
            raise ValueError(f"Mathematical operation failed: {str(e)}") from e

    if isinstance(node, ast.Compare):
        left = evaluate_mathematical_node(node.left, session_vars)
        op = COMPARISON_OPS[type(node.ops[0])]
        right = evaluate_mathematical_node(node.comparators[0], session_vars)
        return op(left, right)

    if isinstance(node, ast.Call):
        func_name = node.func.id
        func = ALL_FUNCTIONS[func_name]
        args = [evaluate_mathematical_node(arg, session_vars) for arg in node.args]

        if func_name == "factorial" and len(args) == 1:
            if args[0] > FACTORIAL_COMPUTATION_UPPER_BOUND:
                raise ValueError(f"Factorial input too large (max: {FACTORIAL_COMPUTATION_UPPER_BOUND})")
            if args[0] < 0 or not isinstance(args[0], int):
                raise ValueError("Factorial requires non-negative integer")

        elif func_name == "pow" and len(args) == 2:
            if abs(args[1]) > EXPONENTIATION_SAFETY_THRESHOLD:
                raise ValueError(f"Power exponent too large (max: {EXPONENTIATION_SAFETY_THRESHOLD})")

        try:
            return func(*args)
        except (OverflowError, ValueError, statistics.StatisticsError) as e:
            raise ValueError(f"Mathematical operation failed: {str(e)}") from e

    if isinstance(node, ast.List):
        return [evaluate_mathematical_node(elt, session_vars) for elt in node.elts]

    raise ValueError(f"Unsupported node type: {type(node).__name__}")


def compute_expression(expression: str, session_id: str | None = None) -> MathematicalComputationResult:
    start_time = time.time()

    try:
        validate_expression_security_constraints(expression)

        expression = expression.strip()
        if not expression:
            raise SyntaxError("Empty expression")

        if len(expression) > MAXIMUM_MATHEMATICAL_EXPRESSION_CHARACTER_LIMIT:
            raise ValueError(f"Expression too long (max: {MAXIMUM_MATHEMATICAL_EXPRESSION_CHARACTER_LIMIT} characters)")

        original_expression = expression

        session_vars = {}
        if session_id:
            session_vars = _session_manager.get_session(session_id) or {}

        cached_result = retrieve_cached_computation_result(original_expression)
        if cached_result is not None:
            return MathematicalComputationResult(
                success=True,
                expression=original_expression,
                result=cached_result,
                metadata={"cache_hit": True, "computation_time": 0},
            )

        expression = sanitize_expression(expression)
        expression = preprocess_expression(expression)

        tree = retrieve_cached_abstract_syntax_tree(expression)
        if tree is None:
            try:
                tree = ast.parse(expression, mode="eval")
                store_parsed_expression(expression, tree)
            except SyntaxError as e:
                raise SyntaxError(f"Invalid mathematical expression: {str(e)}") from e

        try:
            if not validate_ast_node(tree, session_vars=session_vars):
                raise ValueError("Expression contains unsupported operations")
        except Exception as e:
            if "unsupported operations" in str(e).lower():
                raise ValueError(f"Unsupported mathematical operation detected: {str(e)}") from e
            raise

        def compute_result(timeout_event=None):
            with mathematical_computation_memory_limit(MAXIMUM_MEMORY_USAGE_MEGABYTES):
                if timeout_event and timeout_event.is_set():
                    raise TimeoutError(f"Computation exceeded {MATHEMATICAL_OPERATION_TIMEOUT_SECONDS}s limit")
                result = evaluate_mathematical_node(tree, session_vars)
                if timeout_event and timeout_event.is_set():
                    raise TimeoutError(f"Computation exceeded {MATHEMATICAL_OPERATION_TIMEOUT_SECONDS}s limit")
                return result

        with mathematical_computation_timeout(MATHEMATICAL_OPERATION_TIMEOUT_SECONDS) as timeout_event:
            result = compute_result(timeout_event)

        if isinstance(result, (int, float, complex)) and abs(result) > 10**200:
            result_magnitude = abs(result)
            if result_magnitude == float("inf"):
                raise ValueError("Result is infinite - consider using smaller values")
            else:
                raise ValueError(
                    f"Result magnitude ({result_magnitude:.2e}) exceeds safe computational limits"
                )

        result_str = format_calculation_output(result, original_expression)

        if len(result_str) > MAXIMUM_COMPUTATION_RESULT_CHARACTER_LIMIT:
            raise ValueError(f"Result too large (max length: {MAXIMUM_COMPUTATION_RESULT_CHARACTER_LIMIT})")

        mathematical_calculation_history.add(original_expression, result_str)
        persist_computation_result_in_cache(original_expression, result_str)

        end_time = time.time()
        track_mathematical_operation_performance(original_expression, start_time, end_time)

        return MathematicalComputationResult(
            success=True,
            expression=original_expression,
            result=result,
            metadata={"computation_time": end_time - start_time, "cache_hit": False},
        )

    except (
        SyntaxError,
        ValueError,
        ZeroDivisionError,
        TypeError,
        AttributeError,
        TimeoutError,
    ) as e:
        return MathematicalComputationResult(
            success=False,
            expression=expression,
            error={"type": type(e).__name__, "message": str(e)},
        )
    except Exception as e:
        return MathematicalComputationResult(
            success=False,
            expression=expression,
            error={"type": "UnexpectedError", "message": f"Calculation error: {str(e)}"},
        )


def execute_mathematical_computation(expression: str, session_id: str | None = None) -> str:
    result = compute_expression(expression, session_id)
    if result.success:
        return format_calculation_output(result.result, expression)
    else:
        error_msg = result.error.get("message", "Unknown error")
        if "division by zero" in error_msg.lower():
            raise ValueError("Cannot divide by zero")
        elif "invalid" in error_msg.lower() or "unsupported" in error_msg.lower():
            raise ValueError(error_msg)
        elif "empty" in error_msg.lower():
            raise SyntaxError(error_msg)
        else:
            raise ValueError(error_msg)


def format_calculation_output(result: Any, original_expression: str) -> str:
    if isinstance(result, bool):
        return str(result)

    if isinstance(result, complex):
        if result.imag == 0:
            return format_calculation_output(result.real, original_expression)
        return str(result)

    if isinstance(result, (list, tuple)):
        return str(result)

    if isinstance(result, (float, int)):
        has_division = "/" in original_expression or "÷" in original_expression
        has_float_operand = "." in original_expression
        has_statistics_func = any(
            func in original_expression
            for func in ["mean", "median", "stdev", "pstdev", "variance", "pvariance"]
        )

        is_integer_result = isinstance(result, int) or (
            isinstance(result, float) and result.is_integer()
        )

        if "nextafter" in original_expression or "ulp" in original_expression:
            return f"{result:.17g}"

        if has_statistics_func and is_integer_result:
            return f"{float(result):.1f}"

        if is_integer_result and not has_division and not has_float_operand:
            return str(int(result))

        if abs(result) < 1e-10 and result != 0:
            return f"{result:.10e}"
        elif abs(result) > 1e10:
            return f"{result:.6e}"
        else:
            return str(round(result, 10))

    return str(result)


def matrix_multiply(A: list[list[float]], B: list[list[float]]) -> list[list[float]]:
    rows_A, cols_A = len(A), len(A[0]) if A else 0
    rows_B, cols_B = len(B), len(B[0]) if B else 0

    if cols_A != rows_B:
        raise ValueError(f"Cannot multiply matrices: {cols_A} != {rows_B}")

    result = [[0 for _ in range(cols_B)] for _ in range(rows_A)]

    for i in range(rows_A):
        for j in range(cols_B):
            for k in range(cols_A):
                result[i][j] += A[i][k] * B[k][j]

    return result


def matrix_determinant(matrix: list[list[float]]) -> float:
    n = len(matrix)
    if n == 0 or n != len(matrix[0]):
        raise ValueError("Matrix must be square")

    if n == 1:
        return matrix[0][0]

    if n == 2:
        return matrix[0][0] * matrix[1][1] - matrix[0][1] * matrix[1][0]

    det = 0
    for j in range(n):
        minor = [row[:j] + row[j + 1 :] for row in matrix[1:]]
        det += ((-1) ** j) * matrix[0][j] * matrix_determinant(minor)

    return det


def matrix_inverse(matrix: list[list[float]]) -> list[list[float]]:
    n = len(matrix)
    det = matrix_determinant(matrix)

    if abs(det) < 1e-10:
        raise ValueError("Matrix is singular (determinant is zero)")

    if n == 2:
        return [
            [matrix[1][1] / det, -matrix[0][1] / det],
            [-matrix[1][0] / det, matrix[0][0] / det],
        ]

    raise ValueError("Matrix inverse only implemented for 2x2 matrices")


def is_prime(n: int) -> bool:
    if n < 2:
        return False
    if n == 2:
        return True
    if n % 2 == 0:
        return False

    return all(n % i != 0 for i in range(3, int(math.sqrt(n)) + 1, 2))


def prime_factors(n: int) -> list[int]:
    factors = []
    d = 2
    while d * d <= n:
        while n % d == 0:
            factors.append(d)
            n //= d
        d += 1
    if n > 1:
        factors.append(n)
    return factors


def convert_unit(value: float, from_unit: str, to_unit: str, unit_type: str) -> float:
    if unit_type not in UNIT_CONVERSIONS:
        raise ValueError(f"Unknown unit type: {unit_type}")

    conversions = UNIT_CONVERSIONS[unit_type]

    if from_unit not in conversions or to_unit not in conversions:
        raise ValueError(f"Unknown unit in {unit_type}: {from_unit} or {to_unit}")

    if unit_type == "temperature":
        kelvin = conversions[from_unit](value)
        if to_unit == "K":
            return kelvin
        elif to_unit == "C":
            return kelvin - 273.15
        elif to_unit == "F":
            return (kelvin - 273.15) * 9 / 5 + 32
    elif unit_type == "fuel_economy":
        if from_unit == "mpg" and to_unit == "L/100km" or from_unit == "L/100km" and to_unit == "mpg":
            return 235.215 / value
        elif from_unit == "mpg_uk" and to_unit == "L/100km" or from_unit == "L/100km" and to_unit == "mpg_uk":
            return 282.481 / value
        elif from_unit == "km/L" and to_unit == "L/100km" or from_unit == "L/100km" and to_unit == "km/L":
            return 100.0 / value
        elif from_unit == "mpg" and to_unit == "mpg_uk":
            return value / 1.20095
        elif from_unit == "mpg_uk" and to_unit == "mpg":
            return value * 1.20095
        elif from_unit == "mpg" and to_unit == "km/L":
            return value * 0.425144
        elif from_unit == "km/L" and to_unit == "mpg":
            return value * 2.35215
        else:
            base_value = value * conversions[from_unit]
            return base_value / conversions[to_unit]
    else:
        base_value = value * conversions[from_unit]
        return base_value / conversions[to_unit]


class UnitConversionHistory:
    def __init__(self, max_size: int = 100):
        self.history: list[dict[str, Any]] = []
        self.max_size = max_size
        self.lock = RLock()

    def add(self, value: float, from_unit: str, to_unit: str, result: float) -> None:
        with self.lock:
            entry = {
                "value": value,
                "from_unit": from_unit,
                "to_unit": to_unit,
                "result": result,
                "timestamp": datetime.datetime.now().isoformat(),
            }
            self.history.append(entry)
            if len(self.history) > self.max_size:
                self.history.pop(0)

    def get_recent(self, limit: int = 10) -> list[dict[str, Any]]:
        with self.lock:
            return self.history[-limit:]

    def clear(self) -> None:
        with self.lock:
            self.history.clear()


conversion_history = UnitConversionHistory()


def resolve_unit_alias(unit: str) -> str:
    unit_lower = unit.lower()
    return UNIT_ALIASES.get(unit_lower, unit)


def detect_unit_type(unit: str) -> str | None:
    unit = resolve_unit_alias(unit)
    for unit_type, units in UNIT_CONVERSIONS.items():
        if unit in units:
            return unit_type
    return None


def format_scientific_notation(value: float, precision: int = 4) -> str:
    if abs(value) < 1e-10 or abs(value) > 1e10:
        return f"{value:.{precision}e}"
    return str(value)


def convert_with_history(value: float, from_unit: str, to_unit: str, precision: int = 10) -> float:
    from_unit = resolve_unit_alias(from_unit)
    to_unit = resolve_unit_alias(to_unit)

    unit_type = detect_unit_type(from_unit)
    if not unit_type:
        raise ValueError(f"Unknown unit: {from_unit}")

    if to_unit not in UNIT_CONVERSIONS[unit_type]:
        raise ValueError(f"Unknown unit: {to_unit}")

    result = convert_unit(value, from_unit, to_unit, unit_type)
    result = round(result, precision)

    conversion_history.add(value, from_unit, to_unit, result)
    return result


def parse_compound_unit(unit_str: str) -> tuple[list[str], list[str]]:
    unit_str = unit_str.replace("·", "*").replace("/", " / ").replace("*", " * ")

    parts = unit_str.split()
    numerator = []
    denominator = []
    current_list = numerator

    for part in parts:
        if part == "/":
            current_list = denominator
        elif part not in ["*", "·"]:
            if "^" in part or part[-1].isdigit():
                base_unit = "".join(c for c in part if not c.isdigit() and c != "^")
                power = "".join(c for c in part if c.isdigit())
                power = int(power) if power else 1
                for _ in range(abs(power)):
                    if power > 0:
                        current_list.append(base_unit)
                    else:
                        (denominator if current_list == numerator else numerator).append(base_unit)
            else:
                current_list.append(part)

    return numerator, denominator


def calculate_percentage(value: float, percentage: float) -> float:
    return value * (percentage / 100)


def calculate_percentage_of(part: float, whole: float) -> float:
    if whole == 0:
        raise ValueError("Cannot calculate percentage of zero")
    return (part / whole) * 100


def calculate_percentage_change(old_value: float, new_value: float) -> float:
    if old_value == 0:
        raise ValueError("Cannot calculate percentage change from zero")
    return ((new_value - old_value) / old_value) * 100


def split_bill(total: float, people: int, tip_percent: float = 0) -> dict[str, float]:
    if people <= 0:
        raise ValueError("Number of people must be positive")

    tip_amount = calculate_percentage(total, tip_percent)
    total_with_tip = total + tip_amount
    per_person = total_with_tip / people

    return {
        "total": total,
        "tip_amount": tip_amount,
        "total_with_tip": total_with_tip,
        "per_person": per_person,
        "people": people,
    }


def calculate_tax(amount: float, tax_rate: float, is_inclusive: bool = False) -> dict[str, float]:
    if is_inclusive:
        base_amount = amount / (1 + tax_rate / 100)
        tax_amount = amount - base_amount
    else:
        tax_amount = calculate_percentage(amount, tax_rate)
        base_amount = amount

    return {
        "base_amount": base_amount,
        "tax_rate": tax_rate,
        "tax_amount": tax_amount,
        "total_amount": base_amount + tax_amount,
    }


def calculate_tip(amount: float, tip_percent: float) -> dict[str, float]:
    tip_amount = calculate_percentage(amount, tip_percent)
    return {
        "base_amount": amount,
        "tip_percent": tip_percent,
        "tip_amount": tip_amount,
        "total_amount": amount + tip_amount,
    }


def calculate_compound_interest(
    principal: float, rate: float, time: float, compounds_per_year: int = 1
) -> dict[str, float]:
    rate_decimal = rate / 100
    amount = principal * (1 + rate_decimal / compounds_per_year) ** (compounds_per_year * time)
    interest = amount - principal

    return {
        "principal": principal,
        "rate": rate,
        "time": time,
        "compounds_per_year": compounds_per_year,
        "interest": interest,
        "amount": amount,
    }


def calculate_simple_interest(principal: float, rate: float, time: float) -> dict[str, float]:
    rate_decimal = rate / 100
    interest = principal * rate_decimal * time
    amount = principal + interest

    return {
        "principal": principal,
        "rate": rate,
        "time": time,
        "interest": interest,
        "amount": amount,
    }


def calculate_loan_payment(
    principal: float, annual_rate: float, years: float, payments_per_year: int = 12
) -> dict[str, float]:
    if annual_rate == 0:
        payment = principal / (years * payments_per_year)
        total_paid = principal
        interest_paid = 0
    else:
        rate = annual_rate / 100 / payments_per_year
        n = years * payments_per_year
        payment = principal * (rate * (1 + rate) ** n) / ((1 + rate) ** n - 1)
        total_paid = payment * n
        interest_paid = total_paid - principal

    return {
        "principal": principal,
        "annual_rate": annual_rate,
        "years": years,
        "payments_per_year": payments_per_year,
        "payment": payment,
        "total_paid": total_paid,
        "interest_paid": interest_paid,
    }


def calculate_discount(original_price: float, discount_percent: float) -> dict[str, float]:
    discount_amount = calculate_percentage(original_price, discount_percent)
    final_price = original_price - discount_amount

    return {
        "original_price": original_price,
        "discount_percent": discount_percent,
        "discount_amount": discount_amount,
        "final_price": final_price,
    }


def calculate_markup(cost: float, markup_percent: float) -> dict[str, float]:
    markup_amount = calculate_percentage(cost, markup_percent)
    selling_price = cost + markup_amount

    return {
        "cost": cost,
        "markup_percent": markup_percent,
        "markup_amount": markup_amount,
        "selling_price": selling_price,
    }


_shutdown_requested = False
_active_mathematical_computations = 0


def setup_graceful_shutdown():
    global _shutdown_requested

    def shutdown_handler(signum, frame):
        del frame
        _shutdown_requested = True
        async_mathematical_logger.log(
            logging.INFO, f"Received shutdown signal {signum}, initiating graceful shutdown"
        )

    import signal

    try:
        signal.signal(signal.SIGTERM, shutdown_handler)
        signal.signal(signal.SIGINT, shutdown_handler)
    except (OSError, ValueError):
        pass


mcp = FastMCP("MCP Math")


@mcp.tool()
async def get_system_metrics() -> str:
    metrics = ["MCP Math System Metrics:"]
    metrics.append("=" * 40)

    with _computation_metrics_lock:
        stats_values = []
        for i in range(_computation_stats.size()):
            key = (
                list(_computation_stats.cache.keys())[i]
                if i < len(_computation_stats.cache)
                else None
            )
            if key:
                stat = _computation_stats.get(key)
                if stat:
                    stats_values.append(stat)

        total_computations = sum(stats["count"] for stats in stats_values)
        total_time = sum(stats["total_time"] for stats in stats_values)
        avg_time = total_time / total_computations if total_computations > 0 else 0

    metrics.append(f"Total Computations: {total_computations}")
    metrics.append(f"Average Computation Time: {avg_time:.3f}s")
    metrics.append(f"Active Computations: {_active_mathematical_computations}")

    with _rate_limiting_lock:
        active_clients = len(mathematical_computation_rate_limiter.requests)
        total_requests = sum(
            len(reqs) for reqs in mathematical_computation_rate_limiter.requests.values()
        )

    metrics.append(f"Active Clients: {active_clients}")
    metrics.append(f"Total Requests (current window): {total_requests}")

    history_size = len(mathematical_calculation_history.history)
    metrics.append(f"History Entries: {history_size}/{CALCULATION_HISTORY_ENTRY_LIMIT}")

    cache_size = _expression_cache.size()
    ast_cache_size = _parsed_expression_ast_cache.size()
    metrics.append(f"Expression Cache: {cache_size}/{MAX_CACHE_SIZE}")
    metrics.append(f"AST Cache: {ast_cache_size}/{MAX_CACHE_SIZE}")

    session_stats = _session_manager.get_stats()
    metrics.append(
        f"Active Sessions: {session_stats['active_sessions']}/{session_stats['max_sessions']}"
    )

    metrics.append(f"Security Monitoring: {'Enabled' if SECURITY_AUDIT_LOGGING_ENABLED else 'Disabled'}")
    metrics.append(f"Rate Limiting: {'Enabled' if ENABLE_RATE_LIMITING else 'Disabled'}")

    metrics.append("\\nConfiguration:")
    metrics.append(f"  Max Expression Length: {MAXIMUM_MATHEMATICAL_EXPRESSION_CHARACTER_LIMIT}")
    metrics.append(f"  Max Expression Depth: {MAXIMUM_AST_NODE_DEPTH_LIMIT}")
    metrics.append(f"  Max Computation Time: {MAXIMUM_COMPUTATION_DURATION_SECONDS}s")
    metrics.append(f"  Computation Timeout: {MATHEMATICAL_OPERATION_TIMEOUT_SECONDS}s")
    metrics.append(f"  Memory Limit: {MAXIMUM_MEMORY_USAGE_MEGABYTES}MB")
    metrics.append(f"  Available Functions: {len(ALL_FUNCTIONS)}")

    return "\\n".join(metrics)


@mcp.tool()
async def get_security_status() -> str:
    security_info = ["MCP Math Security Status:"]
    security_info.append("=" * 35)

    security_info.append("Rate Limiting:")
    security_info.append(f"  Status: {'Enabled' if ENABLE_RATE_LIMITING else 'Disabled'}")
    security_info.append(f"  Window: {RATE_LIMIT_WINDOW}s")
    security_info.append(f"  Max Requests: {MAXIMUM_CLIENT_REQUESTS_PER_TIME_WINDOW}")

    with _rate_limiting_lock:
        security_info.append(
            f"  Active Clients: {len(mathematical_computation_rate_limiter.requests)}"
        )

    security_info.append("\\nThreat Detection:")
    security_info.append(f"  Forbidden Patterns: {len(FORBIDDEN_PATTERNS)}")
    security_info.append(f"  Input Hashing: {'Enabled' if ENABLE_INPUT_HASHING else 'Disabled'}")
    security_info.append(f"  Timeout Protection: {MATHEMATICAL_OPERATION_TIMEOUT_SECONDS}s")
    security_info.append(f"  Memory Protection: {MAXIMUM_MEMORY_USAGE_MEGABYTES}MB")

    security_info.append("\\nAudit Configuration:")
    security_info.append(f"  Audit Logging: {'Enabled' if SECURITY_AUDIT_LOGGING_ENABLED else 'Disabled'}")
    security_info.append("  Security Logger: Active")
    security_info.append("  Async Logging: Enabled")

    security_info.append("\\nResource Limits:")
    security_info.append(f"  Max Factorial: {FACTORIAL_COMPUTATION_UPPER_BOUND}")
    security_info.append(f"  Max Power Exponent: {EXPONENTIATION_SAFETY_THRESHOLD}")
    security_info.append(f"  Max Result Length: {MAXIMUM_COMPUTATION_RESULT_CHARACTER_LIMIT}")
    security_info.append(f"  Max List Size: {MAXIMUM_LIST_ELEMENT_COUNT_LIMIT}")

    return "\\n".join(security_info)


@mcp.tool()
async def get_memory_usage() -> str:
    try:
        stats = get_memory_usage_stats()
        lines = ["MCP Math Memory Usage:"]
        lines.append("=" * 30)
        lines.append(f"Process Memory: {stats['process_memory_mb']} MB")
        lines.append(f"Virtual Memory: {stats['virtual_memory_mb']} MB")
        lines.append("\\nCache Usage:")
        lines.append(f"  Expression Cache: {stats['expression_cache_size']}/{MAX_CACHE_SIZE}")
        lines.append(f"  AST Cache: {stats['ast_cache_size']}/{MAX_CACHE_SIZE}")
        lines.append(
            f"  Computation Stats: {stats['computation_stats_size']}/{MAX_COMPUTATION_STATS}"
        )
        lines.append("\\nSession Management:")
        lines.append(f"  Active Sessions: {stats['active_sessions']}/{MAX_SESSIONS}")
        lines.append("\\nOther Components:")
        lines.append(f"  Rate Limiter Clients: {stats['rate_limiter_clients']}")
        lines.append(f"  History Entries: {stats['history_entries']}/{CALCULATION_HISTORY_ENTRY_LIMIT}")
        return "\\n".join(lines)
    except ImportError:
        return "Error: psutil package required for memory monitoring"
    except Exception as e:
        return f"Error getting memory usage: {str(e)}"


@mcp.tool()
async def calculate(expression: str) -> str:
    global _active_mathematical_computations

    if _shutdown_requested:
        return "Error: Server is shutting down, not accepting new requests"

    _active_mathematical_computations += 1

    try:
        result = compute_expression(expression)
        return result.to_string()
    except Exception as e:
        return f"Error: {str(e)}"
    finally:
        _active_mathematical_computations -= 1


@mcp.tool()
async def batch_calculate(expressions: list[str]) -> str:
    results = []
    for expr in expressions:
        result = compute_expression(expr)
        if result.success:
            results.append(f"{expr} = {result.result}")
        else:
            results.append(f"{expr} = Error: {result.error['message']}")
    return "\\n".join(results)


@mcp.tool()
async def statistics(data: list[float], operation: str) -> str:
    try:
        if operation not in STATISTICS_FUNCTIONS:
            return f"Error: Unknown operation {operation}"

        func = STATISTICS_FUNCTIONS[operation]
        result = func(data)

        return f"Result: {result}"
    except statistics.StatisticsError as e:
        return f"Error: {str(e)}"
    except Exception as e:
        return f"Error: {str(e)}"


@mcp.tool()
async def matrix_operation(matrices: list[list[list[float]]], operation: str) -> str:
    try:
        if operation == "multiply":
            if len(matrices) != 2:
                return "Error: Matrix multiply requires exactly 2 matrices"
            result = matrix_multiply(matrices[0], matrices[1])
            return f"Result: {result}"

        elif operation == "determinant":
            if len(matrices) != 1:
                return "Error: Determinant requires exactly 1 matrix"
            result = matrix_determinant(matrices[0])
            return f"Result: {result}"

        elif operation == "inverse":
            if len(matrices) != 1:
                return "Error: Inverse requires exactly 1 matrix"
            result = matrix_inverse(matrices[0])
            return f"Result: {result}"

        else:
            return f"Error: Unknown operation {operation}"

    except Exception as e:
        return f"Error: {str(e)}"


@mcp.tool()
async def convert_units(value: float, from_unit: str, to_unit: str, unit_type: str) -> str:
    try:
        result = convert_unit(value, from_unit, to_unit, unit_type)
        return f"Result: {value} {from_unit} = {result} {to_unit}"
    except Exception as e:
        return f"Error: {str(e)}"


@mcp.tool()
async def number_theory(n: int, operation: str) -> str:
    try:
        if operation == "is_prime":
            result = is_prime(n)
            return f"Result: {n} is {'prime' if result else 'not prime'}"

        elif operation == "prime_factors":
            result = prime_factors(n)
            return f"Result: Prime factors of {n} = {result}"

        elif operation == "divisors":
            divisors = [i for i in range(1, n + 1) if n % i == 0]
            return f"Result: Divisors of {n} = {divisors}"

        elif operation == "totient":
            result = sum(1 for i in range(1, n) if math.gcd(i, n) == 1)
            return f"Result: Euler's totient of {n} = {result}"

        else:
            return f"Error: Unknown operation {operation}"

    except Exception as e:
        return f"Error: {str(e)}"


@mcp.tool()
async def create_session(
    session_id: str | None = None, variables: dict[str, float] | None = None
) -> str:
    try:
        if not session_id:
            session_id = f"session_{int(time.time() * 1000)}"

        _session_manager.create_session(session_id, variables)
        return f"Session created: {session_id}"
    except Exception as e:
        return f"Error: {str(e)}"


@mcp.tool()
async def calculate_in_session(session_id: str, expression: str, save_as: str | None = None) -> str:
    global _active_mathematical_computations

    if _shutdown_requested:
        return "Error: Server is shutting down"

    _active_mathematical_computations += 1

    try:
        result = compute_expression(expression, session_id)

        if result.success and save_as:
            _session_manager.update_session(session_id, save_as, result.result)

        return result.to_string()
    except Exception as e:
        return f"Error: {str(e)}"
    finally:
        _active_mathematical_computations -= 1


@mcp.tool()
async def list_session_variables(session_id: str) -> str:
    variables = _session_manager.get_session(session_id)
    if variables is None:
        return f"Error: Session {session_id} not found"

    if not variables:
        return "No variables in session"

    lines = ["Session Variables:"]
    for name, value in variables.items():
        lines.append(f"  {name} = {value}")

    return "\\n".join(lines)


@mcp.tool()
async def delete_session(session_id: str) -> str:
    if _session_manager.delete_session(session_id):
        return f"Session {session_id} deleted"
    return f"Session {session_id} not found"


@mcp.tool()
async def get_calculation_history(limit: int = 10) -> str:
    if limit > CALCULATION_HISTORY_ENTRY_LIMIT:
        limit = CALCULATION_HISTORY_ENTRY_LIMIT

    history = mathematical_calculation_history.get_recent(limit)
    if not history:
        return "No calculation history available."

    lines = ["Recent Calculations:"]
    for entry in history:
        lines.append(f"  {entry['expression']} = {entry['result']} ({entry['timestamp']})")
    return "\\n".join(lines)


@mcp.tool()
async def clear_history() -> str:
    mathematical_calculation_history.clear()
    return "Calculation history cleared successfully"


@mcp.tool()
async def cleanup_memory() -> str:
    try:
        cleanup_stats = cleanup_expired_cache_entries()
        lines = ["Memory Cleanup Results:"]
        lines.append(f"Expression cache expired: {cleanup_stats['expression_cache_expired']}")
        lines.append(f"Sessions expired: {cleanup_stats['sessions_expired']}")
        lines.append(f"Rate limiter expired: {cleanup_stats['rate_limiter_expired']}")
        total_cleaned = sum(cleanup_stats.values())
        lines.append(f"Total items cleaned: {total_cleaned}")
        return "\\n".join(lines)
    except Exception as e:
        return f"Error during cleanup: {str(e)}"


@mcp.tool()
async def list_functions() -> str:
    lines = ["Available Mathematical Functions and Constants:"]

    lines.append("\\nBasic Math Functions:")
    basic_funcs = ["sin", "cos", "tan", "sqrt", "exp", "log", "log10", "log2", "pow", "factorial"]
    for func in sorted(f for f in MATH_FUNCTIONS if f in basic_funcs):
        lines.append(f"  {func}")

    lines.append("\\nStatistics Functions:")
    for func in sorted(STATISTICS_FUNCTIONS.keys()):
        lines.append(f"  {func}")

    lines.append("\\nComplex Number Functions:")
    for func in sorted(COMPLEX_FUNCTIONS.keys()):
        lines.append(f"  {func}")

    lines.append("\\nConstants:")
    for const in sorted(MATH_CONSTANTS.keys()):
        lines.append(f"  {const} = {MATH_CONSTANTS[const]}")

    lines.append("\\nOperators:")
    lines.append("  +, -, *, /, //, %, **, ×, ÷, ^")
    lines.append("  ==, !=, <, <=, >, >=")

    lines.append("\\nUnit Types:")
    for unit_type in UNIT_CONVERSIONS:
        lines.append(f"  {unit_type}: {', '.join(UNIT_CONVERSIONS[unit_type].keys())}")

    return "\\n".join(lines)


@mcp.resource("history://recent")
async def get_recent_history() -> str:
    history = mathematical_calculation_history.get_recent(20)
    if not history:
        return "No calculation history available."

    lines = ["Recent Calculations:"]
    for entry in history:
        lines.append(f"  {entry['expression']} = {entry['result']}")
    return "\\n".join(lines)


@mcp.resource("functions://available")
async def get_available_functions() -> str:
    lines = ["Available Mathematical Functions:"]

    lines.append("\\nTrigonometric:")
    for func in ["sin", "cos", "tan", "asin", "acos", "atan"]:
        if func in MATH_FUNCTIONS:
            lines.append(f"  {func}")

    lines.append("\\nHyperbolic:")
    for func in ["sinh", "cosh", "tanh", "asinh", "acosh", "atanh"]:
        if func in MATH_FUNCTIONS:
            lines.append(f"  {func}")

    lines.append("\\nLogarithmic:")
    for func in ["log", "log10", "log2", "exp"]:
        if func in MATH_FUNCTIONS:
            lines.append(f"  {func}")

    lines.append("\\nStatistical:")
    for func in sorted(STATISTICS_FUNCTIONS.keys()):
        lines.append(f"  {func}")

    lines.append("\\nOther:")
    other_funcs = sorted(
        set(MATH_FUNCTIONS.keys())
        - {
            "sin",
            "cos",
            "tan",
            "asin",
            "acos",
            "atan",
            "sinh",
            "cosh",
            "tanh",
            "asinh",
            "acosh",
            "atanh",
            "log",
            "log10",
            "log2",
            "exp",
        }
    )
    for func in other_funcs:
        lines.append(f"  {func}")

    return "\\n".join(lines)


@mcp.resource("constants://math")
async def get_math_constants() -> str:
    lines = ["Mathematical Constants:"]
    for name, value in MATH_CONSTANTS.items():
        lines.append(f"  {name} = {value}")
    return "\\n".join(lines)


@mcp.prompt()
async def scientific_calculation() -> str:
    return """I need help with scientific calculations. Here are some examples:

Basic Operations:
- Addition: 2 + 3
- Subtraction: 10 - 4
- Multiplication: 5 * 6 or 5 × 6
- Division: 15 / 3 or 15 ÷ 3
- Power: 2 ** 3 or 2 ^ 3

Scientific Functions:
- Trigonometry: sin(pi/2), cos(0), tan(pi/4)
- Logarithms: log(10), log10(100), log2(8)
- Square root: sqrt(16)
- Exponential: exp(1)

Statistics:
- mean([1,2,3,4,5])
- stdev([1,2,3,4,5])

What calculation would you like me to perform?"""


@mcp.prompt()
async def batch_calculation() -> str:
    return """I need to process multiple calculations at once.

Example: ["2 + 2", "sin(pi/2)", "sqrt(16) * cos(0)", "factorial(5)", "log10(1000)"]

What expressions would you like to calculate?"""


def main() -> None:
    setup_graceful_shutdown()
    start_memory_cleanup_timer()

    if SECURITY_AUDIT_LOGGING_ENABLED:
        async_mathematical_logger.log(
            logging.INFO, "Starting Enhanced MCP Math server with memory management"
        )
        async_mathematical_logger.log(logging.INFO, f"Available functions: {len(ALL_FUNCTIONS)}")
        async_mathematical_logger.log(
            logging.INFO, f"Security: timeout={MATHEMATICAL_OPERATION_TIMEOUT_SECONDS}s, memory={MAXIMUM_MEMORY_USAGE_MEGABYTES}MB"
        )
        async_mathematical_logger.log(
            logging.INFO,
            f"Memory: max_cache={MAX_CACHE_SIZE}, max_sessions={MAX_SESSIONS}, cleanup_interval={SESSION_CLEANUP_INTERVAL}s",
        )

    try:
        mcp.run()
    except Exception as e:
        if SECURITY_AUDIT_LOGGING_ENABLED:
            async_mathematical_logger.log(
                logging.CRITICAL, f"Failed to start MCP Math server: {str(e)}"
            )
        raise
    finally:
        shutdown_memory_management()


if __name__ == "__main__":
    main()



import math
import unittest

from src.mcp_mathematics.calculator import execute_mathematical_computation


class TestCalculator(unittest.TestCase):
    def test_sin(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("sin(0)")), 0.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("sin(pi/2)")), 1.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("sin(pi)")), 0.0, places=7)

    def test_cos(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("cos(0)")), 1.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("cos(pi/2)")), 0.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("cos(pi)")), -1.0, places=7)

    def test_tan(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("tan(0)")), 0.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("tan(pi/4)")), 1.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("tan(-pi/4)")), -1.0, places=7
        )

    def test_asin(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("asin(0)")), 0.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("asin(1)")), math.pi / 2, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("asin(0.5)")), math.pi / 6, places=7
        )

    def test_acos(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("acos(1)")), 0.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("acos(0)")), math.pi / 2, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("acos(-1)")), math.pi, places=7
        )

    def test_atan(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("atan(0)")), 0.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("atan(1)")), math.pi / 4, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("atan(-1)")), -math.pi / 4, places=7
        )

    def test_atan2(self):
        self.assertAlmostEqual(
            float(execute_mathematical_computation("atan2(0, 1)")), 0.0, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("atan2(1, 0)")), math.pi / 2, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("atan2(1, 1)")), math.pi / 4, places=7
        )

    def test_sinh(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("sinh(0)")), 0.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("sinh(1)")), math.sinh(1), places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("sinh(-1)")), math.sinh(-1), places=7
        )

    def test_cosh(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("cosh(0)")), 1.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("cosh(1)")), math.cosh(1), places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("cosh(-1)")), math.cosh(-1), places=7
        )

    def test_tanh(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("tanh(0)")), 0.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("tanh(1)")), math.tanh(1), places=7
        )
        self.assertAlmostEqual(float(execute_mathematical_computation("tanh(100)")), 1.0, places=7)

    def test_asinh(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("asinh(0)")), 0.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("asinh(1)")), math.asinh(1), places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("asinh(-1)")), math.asinh(-1), places=7
        )

    def test_acosh(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("acosh(1)")), 0.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("acosh(2)")), math.acosh(2), places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("acosh(10)")), math.acosh(10), places=7
        )

    def test_atanh(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("atanh(0)")), 0.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("atanh(0.5)")), math.atanh(0.5), places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("atanh(-0.5)")), math.atanh(-0.5), places=7
        )

    def test_log(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("log(1)")), 0.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("log(e)")), 1.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("log(10)")), math.log(10), places=7
        )

    def test_log10(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("log10(1)")), 0.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("log10(10)")), 1.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("log10(100)")), 2.0, places=7)

    def test_log2(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("log2(1)")), 0.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("log2(2)")), 1.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("log2(8)")), 3.0, places=7)

    def test_log1p(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("log1p(0)")), 0.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("log1p(1)")), math.log(2), places=7
        )
        self.assertAlmostEqual(float(execute_mathematical_computation("log1p(e-1)")), 1.0, places=7)

    def test_exp(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("exp(0)")), 1.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("exp(1)")), math.e, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("exp(2)")), math.exp(2), places=7
        )

    def test_exp2(self):
        try:
            self.assertAlmostEqual(
                float(execute_mathematical_computation("exp2(0)")), 1.0, places=7
            )
            self.assertAlmostEqual(
                float(execute_mathematical_computation("exp2(1)")), 2.0, places=7
            )
            self.assertAlmostEqual(
                float(execute_mathematical_computation("exp2(3)")), 8.0, places=7
            )
        except Exception:
            self.skipTest("exp2 not available in this Python version")

    def test_expm1(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("expm1(0)")), 0.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("expm1(1)")), math.e - 1, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("expm1(-1)")), math.expm1(-1), places=7
        )

    def test_sqrt(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("sqrt(0)")), 0.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("sqrt(4)")), 2.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("sqrt(9)")), 3.0, places=7)

    def test_cbrt(self):
        try:
            self.assertAlmostEqual(
                float(execute_mathematical_computation("cbrt(0)")), 0.0, places=7
            )
            self.assertAlmostEqual(
                float(execute_mathematical_computation("cbrt(8)")), 2.0, places=7
            )
            self.assertAlmostEqual(
                float(execute_mathematical_computation("cbrt(27)")), 3.0, places=7
            )
        except Exception:
            self.skipTest("cbrt not available in this Python version")

    def test_pow(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("pow(2, 0)")), 1.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("pow(2, 3)")), 8.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("pow(5, 2)")), 25.0, places=7)

    def test_hypot(self):
        self.assertAlmostEqual(
            float(execute_mathematical_computation("hypot(3, 4)")), 5.0, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("hypot(5, 12)")), 13.0, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("hypot(0, 0)")), 0.0, places=7
        )

    def test_fabs(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("fabs(0)")), 0.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("fabs(-5)")), 5.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("fabs(3.14)")), 3.14, places=7
        )

    def test_copysign(self):
        self.assertAlmostEqual(
            float(execute_mathematical_computation("copysign(1, -1)")), -1.0, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("copysign(-5, 1)")), 5.0, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("copysign(3.14, -1)")), -3.14, places=7
        )

    def test_factorial(self):
        self.assertEqual(execute_mathematical_computation("factorial(0)"), "1")
        self.assertEqual(execute_mathematical_computation("factorial(5)"), "120")
        self.assertEqual(execute_mathematical_computation("factorial(10)"), "3628800")

    def test_ceil(self):
        self.assertEqual(execute_mathematical_computation("ceil(1.1)"), "2")
        self.assertEqual(execute_mathematical_computation("ceil(2.9)"), "3")
        self.assertEqual(execute_mathematical_computation("ceil(-1.1)"), "-1")

    def test_floor(self):
        self.assertEqual(execute_mathematical_computation("floor(1.9)"), "1")
        self.assertEqual(execute_mathematical_computation("floor(2.1)"), "2")
        self.assertEqual(execute_mathematical_computation("floor(-1.1)"), "-2")

    def test_trunc(self):
        self.assertEqual(execute_mathematical_computation("trunc(1.9)"), "1")
        self.assertEqual(execute_mathematical_computation("trunc(-1.9)"), "-1")
        self.assertEqual(execute_mathematical_computation("trunc(3.5)"), "3")

    def test_degrees(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("degrees(0)")), 0.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("degrees(pi)")), 180.0, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("degrees(pi/2)")), 90.0, places=7
        )

    def test_radians(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("radians(0)")), 0.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("radians(180)")), math.pi, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("radians(90)")), math.pi / 2, places=7
        )

    def test_gcd(self):
        self.assertEqual(execute_mathematical_computation("gcd(12, 8)"), "4")
        self.assertEqual(execute_mathematical_computation("gcd(15, 25)"), "5")
        self.assertEqual(execute_mathematical_computation("gcd(17, 19)"), "1")

    def test_lcm(self):
        try:
            self.assertEqual(execute_mathematical_computation("lcm(4, 6)"), "12")
            self.assertEqual(execute_mathematical_computation("lcm(3, 5)"), "15")
            self.assertEqual(execute_mathematical_computation("lcm(12, 18)"), "36")
        except Exception:
            self.skipTest("lcm not available in this Python version")

    def test_isqrt(self):
        try:
            self.assertEqual(execute_mathematical_computation("isqrt(0)"), "0")
            self.assertEqual(execute_mathematical_computation("isqrt(4)"), "2")
            self.assertEqual(execute_mathematical_computation("isqrt(10)"), "3")
        except Exception:
            self.skipTest("isqrt not available in this Python version")

    def test_fmod(self):
        self.assertAlmostEqual(
            float(execute_mathematical_computation("fmod(10, 3)")), 1.0, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("fmod(5.5, 2.5)")), 0.5, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("fmod(-10, 3)")), -1.0, places=7
        )

    def test_remainder(self):
        self.assertAlmostEqual(
            float(execute_mathematical_computation("remainder(10, 3)")), 1.0, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("remainder(10, 6)")), -2.0, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("remainder(10, 10)")), 0.0, places=7
        )

    def test_modf(self):
        result = execute_mathematical_computation("modf(1.5)")
        self.assertIn("0.5", result)
        self.assertIn("1.0", result)
        result = execute_mathematical_computation("modf(2.7)")
        self.assertIn("0.7", result)
        result = execute_mathematical_computation("modf(-1.5)")
        self.assertIn("-0.5", result)

    def test_frexp(self):
        result = execute_mathematical_computation("frexp(4)")
        self.assertIn("0.5", result)
        self.assertIn("3", result)
        result = execute_mathematical_computation("frexp(8)")
        self.assertIn("0.5", result)
        self.assertIn("4", result)
        result = execute_mathematical_computation("frexp(1)")
        self.assertIn("0.5", result)

    def test_ldexp(self):
        self.assertAlmostEqual(
            float(execute_mathematical_computation("ldexp(0.5, 3)")), 4.0, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("ldexp(1, 0)")), 1.0, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("ldexp(0.5, 4)")), 8.0, places=7
        )

    def test_isfinite(self):
        self.assertEqual(execute_mathematical_computation("isfinite(0)"), "True")
        self.assertEqual(execute_mathematical_computation("isfinite(1.5)"), "True")
        self.assertEqual(execute_mathematical_computation("isfinite(inf)"), "False")

    def test_isinf(self):
        self.assertEqual(execute_mathematical_computation("isinf(inf)"), "True")
        self.assertEqual(execute_mathematical_computation("isinf(-inf)"), "True")
        self.assertEqual(execute_mathematical_computation("isinf(1.0)"), "False")

    def test_isnan(self):
        self.assertEqual(execute_mathematical_computation("isnan(nan)"), "True")
        self.assertEqual(execute_mathematical_computation("isnan(1.0)"), "False")
        self.assertEqual(execute_mathematical_computation("isnan(inf)"), "False")

    def test_isclose(self):
        self.assertEqual(execute_mathematical_computation("isclose(1.0, 1.0)"), "True")
        self.assertEqual(execute_mathematical_computation("isclose(1.0, 1.00001)"), "False")
        self.assertEqual(execute_mathematical_computation("isclose(1.0, 1.000001)"), "False")

    def test_comb(self):
        try:
            self.assertEqual(execute_mathematical_computation("comb(5, 2)"), "10")
            self.assertEqual(execute_mathematical_computation("comb(10, 3)"), "120")
            self.assertEqual(execute_mathematical_computation("comb(4, 4)"), "1")
        except Exception:
            self.skipTest("comb not available in this Python version")

    def test_perm(self):
        try:
            self.assertEqual(execute_mathematical_computation("perm(5, 2)"), "20")
            self.assertEqual(execute_mathematical_computation("perm(4, 3)"), "24")
            self.assertEqual(execute_mathematical_computation("perm(3, 3)"), "6")
        except Exception:
            self.skipTest("perm not available in this Python version")

    def test_erf(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("erf(0)")), 0.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("erf(1)")), math.erf(1), places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("erf(-1)")), math.erf(-1), places=7
        )

    def test_erfc(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("erfc(0)")), 1.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("erfc(1)")), math.erfc(1), places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("erfc(-1)")), math.erfc(-1), places=7
        )

    def test_gamma(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("gamma(1)")), 1.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("gamma(2)")), 1.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("gamma(5)")), 24.0, places=7)

    def test_lgamma(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("lgamma(1)")), 0.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("lgamma(2)")), 0.0, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("lgamma(10)")), math.lgamma(10), places=7
        )

    def test_nextafter(self):
        try:
            result = float(execute_mathematical_computation("nextafter(1, 2)"))
            self.assertGreater(result, 1.0)
            self.assertLess(result, 1.0001)
            result = float(execute_mathematical_computation("nextafter(1, 0)"))
            self.assertLess(result, 1.0)
            self.assertAlmostEqual(
                float(execute_mathematical_computation("nextafter(0, 1)")), 5e-324, places=320
            )
        except Exception:
            self.skipTest("nextafter not available in this Python version")

    def test_ulp(self):
        try:
            self.assertGreater(float(execute_mathematical_computation("ulp(1.0)")), 0)
            self.assertLess(float(execute_mathematical_computation("ulp(1.0)")), 1e-10)
            self.assertGreater(
                float(execute_mathematical_computation("ulp(1000000.0)")),
                float(execute_mathematical_computation("ulp(1.0)")),
            )
        except Exception:
            self.skipTest("ulp not available in this Python version")

    def test_pi_constant(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("pi")), math.pi, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("pi * 2")), math.pi * 2, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("pi / 2")), math.pi / 2, places=7
        )

    def test_e_constant(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("e")), math.e, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("e * 2")), math.e * 2, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("e ** 2")), math.e**2, places=7
        )

    def test_tau_constant(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("tau")), math.tau, places=7)
        self.assertAlmostEqual(
            float(execute_mathematical_computation("tau / 4")), math.tau / 4, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("tau / 2")), math.pi, places=7
        )

    def test_inf_constant(self):
        with self.assertRaises(ValueError, msg="Calculator should handle inf safely"):
            execute_mathematical_computation("inf")
        with self.assertRaises(ValueError):
            execute_mathematical_computation("1.0 / 0.0")
        try:
            result = execute_mathematical_computation("10**308")
            self.assertIsInstance(result, str)
        except ValueError:
            pass

    def test_nan_constant(self):
        self.assertEqual(execute_mathematical_computation("nan"), "nan")
        self.assertEqual(execute_mathematical_computation("nan + 1"), "nan")
        self.assertEqual(execute_mathematical_computation("0 * nan"), "nan")

    def test_basic_arithmetic(self):
        self.assertEqual(execute_mathematical_computation("2 + 3"), "5")
        self.assertEqual(execute_mathematical_computation("10 - 4"), "6")
        self.assertEqual(execute_mathematical_computation("5 * 6"), "30")
        self.assertEqual(execute_mathematical_computation("15 / 3"), "5.0")
        self.assertEqual(execute_mathematical_computation("7 // 2"), "3")
        self.assertEqual(execute_mathematical_computation("10 % 3"), "1")
        self.assertEqual(execute_mathematical_computation("2 ** 3"), "8")

        self.assertEqual(execute_mathematical_computation("98765 + 87654"), "186419")
        self.assertEqual(execute_mathematical_computation("999999 - 123456"), "876543")
        self.assertEqual(execute_mathematical_computation("12345 * 6789"), "83810205")
        self.assertAlmostEqual(
            float(execute_mathematical_computation("987654321 / 123")), 8029709.927, places=2
        )
        self.assertEqual(execute_mathematical_computation("54321 // 789"), "68")
        self.assertEqual(execute_mathematical_computation("987654 % 12345"), "54")
        self.assertEqual(execute_mathematical_computation("123 ** 4"), "228886641")

    def test_unicode_operators(self):
        self.assertEqual(execute_mathematical_computation("2 × 3"), "6")
        self.assertEqual(execute_mathematical_computation("8 ÷ 2"), "4.0")
        self.assertEqual(execute_mathematical_computation("2 ^ 3"), "8")

    def test_complex_expressions(self):
        self.assertEqual(execute_mathematical_computation("(2 + 3) * 4"), "20")
        self.assertAlmostEqual(
            float(execute_mathematical_computation("sin(pi/2) + cos(0)")), 2.0, places=7
        )
        result = execute_mathematical_computation("factorial(5) + sqrt(16)")
        self.assertIn(result, ["124", "124.0"])

        self.assertEqual(
            execute_mathematical_computation("(98765 + 12345) * (54321 - 43210)"), "1234543210"
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("((999 * 888) + 777) / 666")), 1333.167, places=2
        )
        self.assertEqual(
            execute_mathematical_computation("(123456 * (789 + 321)) / (987 - 654)"), "411520.0"
        )

    def test_error_handling(self):
        with self.assertRaises(ValueError):
            execute_mathematical_computation("1 / 0")
        with self.assertRaises(SyntaxError):
            execute_mathematical_computation("")
        with self.assertRaises(ValueError):
            execute_mathematical_computation("unknown_func()")

    def test_new_constants_phi_euler(self):
        self.assertAlmostEqual(
            float(execute_mathematical_computation("phi")), (1 + math.sqrt(5)) / 2, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("euler")), 0.5772156649, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("phi * 2")), (1 + math.sqrt(5)), places=7
        )

    def test_statistics_functions(self):
        try:
            result = execute_mathematical_computation("mean([1, 2, 3, 4, 5])")
            if "Error" in result or "not" in result.lower():
                self.skipTest("List syntax not supported in calculator")
            else:
                self.assertEqual(result, "3.0")
        except Exception as e:
            self.skipTest(f"Statistics functions syntax not supported: {e}")

    def test_complex_functions_phase_polar(self):
        self.assertAlmostEqual(
            float(execute_mathematical_computation("phase(1+1j)")), math.pi / 4, places=7
        )
        result = execute_mathematical_computation("polar(1+1j)")
        self.assertTrue("1.414" in result or "Result:" in result)

    def test_complex_trigonometry_functions(self):
        result = execute_mathematical_computation("csin(0)")
        self.assertTrue("0" in result or "Result:" in result)
        result = execute_mathematical_computation("ccos(0)")
        self.assertTrue("1" in result or "Result:" in result)
        result = execute_mathematical_computation("cexp(0)")
        self.assertTrue("1" in result or "Result:" in result)

    def test_complex_logarithm_functions(self):
        result = execute_mathematical_computation("clog(1)")
        self.assertTrue("0" in result or "Result:" in result)
        result = execute_mathematical_computation("clog10(1)")
        self.assertTrue("0" in result or "Result:" in result)
        result = execute_mathematical_computation("csqrt(4)")
        self.assertTrue("2" in result or "Result:" in result)

    def test_enhanced_mathematical_functions(self):
        try:
            result = execute_mathematical_computation("cbrt(27)")
            self.assertAlmostEqual(float(result), 3.0, places=7)
        except Exception:
            self.skipTest("cbrt not available in this Python version")

        try:
            result = execute_mathematical_computation("comb(5, 2)")
            self.assertEqual(result, "10")
        except Exception:
            self.skipTest("comb not available in this Python version")

        try:
            result = execute_mathematical_computation("perm(5, 2)")
            self.assertEqual(result, "20")
        except Exception:
            self.skipTest("perm not available in this Python version")

    def test_complex_number_expressions(self):
        result = execute_mathematical_computation("1+2j")
        self.assertTrue("1" in result and "2j" in result)

        result = execute_mathematical_computation("(1+2j) + (3+4j)")
        self.assertTrue("4" in result and "6j" in result)

        try:
            result = execute_mathematical_computation("abs(3+4j)")
            self.assertAlmostEqual(float(result), 5.0, places=7)
        except Exception:
            self.skipTest("Complex abs function not implemented")

    def test_security_validation(self):
        dangerous_expressions = [
            "__import__",
            "exec()",
            "eval()",
            "globals()",
            "locals()",
            "open()",
            "file()",
        ]

        for expr in dangerous_expressions:
            with self.assertRaises((ValueError, SyntaxError)) as context:
                execute_mathematical_computation(expr)
            error_message = str(context.exception).lower()
            self.assertTrue(
                "forbidden pattern" in error_message
                or "invalid mathematical expression" in error_message
                or "unsupported operations" in error_message,
                f"Expected security error for '{expr}', got: {error_message}",
            )

    def test_resource_limits_factorial(self):
        result = execute_mathematical_computation("factorial(5)")
        self.assertEqual(result, "120")

        result = execute_mathematical_computation("factorial(10)")
        self.assertEqual(result, "3628800")

        try:
            result = execute_mathematical_computation("factorial(200)")
            if "Error" not in result:
                self.fail("Large factorial should be limited")
        except Exception:
            pass

    def test_resource_limits_expression_length(self):
        very_long_expr = "1+" * 2000 + "1"
        try:
            result = execute_mathematical_computation(very_long_expr)
            if "Error" not in result:
                self.fail("Very long expression should be limited")
        except Exception:
            pass

    def test_enhanced_error_handling_complex(self):
        try:
            result = execute_mathematical_computation("log(-1)")
            if "Error" in result or "nan" in result or "inf" in result:
                pass
            else:
                self.fail("log(-1) should produce error or special value")
        except Exception:
            pass

        try:
            result = execute_mathematical_computation("sqrt(-1)")
            if "Error" in result or "j" in result:
                pass
            else:
                self.fail("sqrt(-1) should produce error or complex number")
        except Exception:
            pass

    def test_trigonometric_edge_cases(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("sin(pi)")), 0.0, places=5)
        self.assertAlmostEqual(float(execute_mathematical_computation("cos(2*pi)")), 1.0, places=7)
        self.assertAlmostEqual(float(execute_mathematical_computation("tan(pi/4)")), 1.0, places=7)

    def test_logarithmic_edge_cases(self):
        self.assertEqual(execute_mathematical_computation("log(1)"), "0")
        self.assertAlmostEqual(float(execute_mathematical_computation("log(e)")), 1.0, places=7)
        self.assertEqual(execute_mathematical_computation("log10(1)"), "0")
        self.assertEqual(execute_mathematical_computation("log10(10)"), "1")

    def test_exponential_edge_cases(self):
        self.assertEqual(execute_mathematical_computation("exp(0)"), "1")
        self.assertAlmostEqual(float(execute_mathematical_computation("exp(1)")), math.e, places=7)
        try:
            self.assertEqual(execute_mathematical_computation("exp2(0)"), "1")
            self.assertEqual(execute_mathematical_computation("exp2(3)"), "8")
        except Exception:
            self.skipTest("exp2 not available in this Python version")

    def test_hyperbolic_functions_edge_cases(self):
        self.assertEqual(execute_mathematical_computation("sinh(0)"), "0")
        self.assertEqual(execute_mathematical_computation("cosh(0)"), "1")
        self.assertEqual(execute_mathematical_computation("tanh(0)"), "0")

    def test_inverse_hyperbolic_functions_edge_cases(self):
        self.assertEqual(execute_mathematical_computation("asinh(0)"), "0")
        self.assertEqual(execute_mathematical_computation("acosh(1)"), "0")
        self.assertEqual(execute_mathematical_computation("atanh(0)"), "0")

    def test_special_values_handling(self):
        with self.assertRaises(ValueError):
            execute_mathematical_computation("inf")
        try:
            result = execute_mathematical_computation("nan")
            self.assertEqual(result, "nan")
        except ValueError:
            pass
        with self.assertRaises(ValueError):
            execute_mathematical_computation("1.0 / 0.0")

    def test_precision_and_rounding(self):
        self.assertAlmostEqual(float(execute_mathematical_computation("pi")), math.pi, places=10)
        self.assertAlmostEqual(float(execute_mathematical_computation("e")), math.e, places=10)
        self.assertAlmostEqual(float(execute_mathematical_computation("tau")), math.tau, places=10)

    def test_mathematical_identities(self):
        self.assertAlmostEqual(
            float(execute_mathematical_computation("sin(pi/2)**2 + cos(pi/2)**2")), 1.0, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("exp(log(5))")), 5.0, places=7
        )
        self.assertAlmostEqual(
            float(execute_mathematical_computation("log(exp(3))")), 3.0, places=7
        )

    def test_large_number_handling(self):
        result = execute_mathematical_computation("2**50")
        self.assertEqual(result, str(2**50))

        result = execute_mathematical_computation("factorial(20)")
        expected = str(math.factorial(20))
        self.assertEqual(result, expected)


class TestSmartFeatures(unittest.TestCase):

    def test_unit_aliases(self):
        from src.mcp_mathematics.calculator import resolve_unit_alias

        self.assertEqual(resolve_unit_alias("meters"), "m")
        self.assertEqual(resolve_unit_alias("kilometres"), "km")
        self.assertEqual(resolve_unit_alias("pounds"), "lb")
        self.assertEqual(resolve_unit_alias("fahrenheit"), "F")
        self.assertEqual(resolve_unit_alias("m"), "m")

    def test_unit_type_detection(self):
        from src.mcp_mathematics.calculator import detect_unit_type

        self.assertEqual(detect_unit_type("m"), "length")
        self.assertEqual(detect_unit_type("meters"), "length")
        self.assertEqual(detect_unit_type("kg"), "mass")
        self.assertEqual(detect_unit_type("s"), "time")
        self.assertEqual(detect_unit_type("C"), "temperature")
        self.assertEqual(detect_unit_type("L"), "volume")
        self.assertIsNone(detect_unit_type("unknown"))

    def test_scientific_notation(self):
        from src.mcp_mathematics.calculator import format_scientific_notation

        self.assertEqual(format_scientific_notation(1.23e-12), "1.2300e-12")
        self.assertEqual(format_scientific_notation(5.67e15), "5.6700e+15")
        self.assertEqual(format_scientific_notation(123.456), "123.456")
        self.assertEqual(format_scientific_notation(1.23e-12, precision=2), "1.23e-12")

    def test_compound_unit_parsing(self):
        from src.mcp_mathematics.calculator import parse_compound_unit

        numerator, denominator = parse_compound_unit("m/s")
        self.assertEqual(numerator, ["m"])
        self.assertEqual(denominator, ["s"])

        numerator, denominator = parse_compound_unit("kg·m/s^2")
        self.assertEqual(numerator, ["kg", "m"])
        self.assertEqual(denominator, ["s", "s"])

    def test_conversion_with_history(self):
        from src.mcp_mathematics.calculator import conversion_history, convert_with_history

        conversion_history.clear()
        result = convert_with_history(100, "meters", "feet", precision=2)
        self.assertAlmostEqual(result, 328.08, places=0)

        history = conversion_history.get_recent(1)
        self.assertEqual(len(history), 1)
        self.assertEqual(history[0]["value"], 100)
        self.assertEqual(history[0]["from_unit"], "m")


class TestFinancialCalculations(unittest.TestCase):

    def test_percentage_calculations(self):
        from src.mcp_mathematics.calculator import (
            calculate_percentage,
            calculate_percentage_change,
            calculate_percentage_of,
        )

        self.assertEqual(calculate_percentage(100, 10), 10)
        self.assertEqual(calculate_percentage(250, 20), 50)
        self.assertEqual(calculate_percentage_of(25, 100), 25)
        self.assertEqual(calculate_percentage_of(50, 200), 25)
        self.assertEqual(calculate_percentage_change(100, 150), 50)
        self.assertEqual(calculate_percentage_change(200, 100), -50)

        self.assertEqual(calculate_percentage(98765, 18.5), 18271.525)
        self.assertEqual(calculate_percentage(1234567, 7.25), 89506.1075)
        self.assertAlmostEqual(calculate_percentage_of(123456, 987654), 12.5, places=1)
        self.assertAlmostEqual(calculate_percentage_of(87654, 999999), 8.765, places=3)
        self.assertAlmostEqual(calculate_percentage_change(54321, 98765), 81.817, places=2)
        self.assertAlmostEqual(calculate_percentage_change(999999, 123456), -87.65, places=2)

        with self.assertRaises(ValueError):
            calculate_percentage_of(10, 0)
        with self.assertRaises(ValueError):
            calculate_percentage_change(0, 100)

    def test_split_bill(self):
        from src.mcp_mathematics.calculator import split_bill

        result = split_bill(100, 4, 15)
        self.assertEqual(result["total"], 100)
        self.assertEqual(result["tip_amount"], 15)
        self.assertEqual(result["total_with_tip"], 115)
        self.assertEqual(result["per_person"], 28.75)

        result = split_bill(12345.67, 8, 18)
        self.assertAlmostEqual(result["total"], 12345.67, places=2)
        self.assertAlmostEqual(result["tip_amount"], 2222.22, places=2)
        self.assertAlmostEqual(result["total_with_tip"], 14567.89, places=2)
        self.assertAlmostEqual(result["per_person"], 1820.99, places=2)

        result = split_bill(98765.43, 15, 20)
        self.assertAlmostEqual(result["tip_amount"], 19753.09, places=2)
        self.assertAlmostEqual(result["total_with_tip"], 118518.52, places=2)
        self.assertAlmostEqual(result["per_person"], 7901.23, places=2)

        with self.assertRaises(ValueError):
            split_bill(100, 0)

    def test_tax_calculations(self):
        from src.mcp_mathematics.calculator import calculate_tax

        result = calculate_tax(100, 10)
        self.assertEqual(result["tax_amount"], 10)
        self.assertEqual(result["total_amount"], 110)

        result = calculate_tax(110, 10, is_inclusive=True)
        self.assertAlmostEqual(result["base_amount"], 100, places=1)
        self.assertAlmostEqual(result["tax_amount"], 10, places=1)

    def test_interest_calculations(self):
        from src.mcp_mathematics.calculator import (
            calculate_compound_interest,
            calculate_simple_interest,
        )

        result = calculate_compound_interest(1000, 5, 10, 12)
        self.assertAlmostEqual(result["amount"], 1647.01, places=0)
        self.assertAlmostEqual(result["interest"], 647.01, places=0)

        result = calculate_simple_interest(1000, 5, 10)
        self.assertEqual(result["interest"], 500)
        self.assertEqual(result["amount"], 1500)

        result = calculate_compound_interest(250000, 7.5, 30, 12)
        self.assertAlmostEqual(result["amount"], 2355383.48, places=2)
        self.assertAlmostEqual(result["interest"], 2105383.48, places=2)

        result = calculate_compound_interest(999999, 12.75, 15, 365)
        self.assertAlmostEqual(result["amount"], 6767725.38, places=2)

        result = calculate_simple_interest(500000, 8.5, 25)
        self.assertEqual(result["interest"], 1062500)
        self.assertEqual(result["amount"], 1562500)

    def test_loan_payment(self):
        from src.mcp_mathematics.calculator import calculate_loan_payment

        result = calculate_loan_payment(100000, 5, 30, 12)
        self.assertAlmostEqual(result["payment"], 536.82, places=0)
        self.assertAlmostEqual(result["total_paid"], 193255.78, delta=10)

        result = calculate_loan_payment(10000, 0, 5)
        self.assertEqual(result["payment"], 166.66666666666666)
        self.assertEqual(result["interest_paid"], 0)

        result = calculate_loan_payment(750000, 6.5, 30, 12)
        self.assertAlmostEqual(result["payment"], 4740.51, places=2)
        self.assertAlmostEqual(result["total_paid"], 1706583.6, delta=100)

        result = calculate_loan_payment(1250000, 8.25, 15, 12)
        self.assertAlmostEqual(result["payment"], 12126.75, places=2)
        self.assertAlmostEqual(result["total_paid"], 2182815.81, delta=100)

        result = calculate_loan_payment(999999, 7.75, 25, 12)
        self.assertAlmostEqual(result["payment"], 7553.28, places=2)

    def test_discount_markup(self):
        from src.mcp_mathematics.calculator import calculate_discount, calculate_markup

        result = calculate_discount(100, 20)
        self.assertEqual(result["discount_amount"], 20)
        self.assertEqual(result["final_price"], 80)

        result = calculate_markup(50, 100)
        self.assertEqual(result["markup_amount"], 50)
        self.assertEqual(result["selling_price"], 100)


class TestUnitConversions(unittest.TestCase):

    def test_length_conversions(self):
        from src.mcp_mathematics.calculator import convert_unit

        self.assertAlmostEqual(convert_unit(1, "m", "km", "length"), 0.001, places=3)
        self.assertAlmostEqual(convert_unit(1, "km", "mi", "length"), 0.621371, places=3)
        self.assertAlmostEqual(convert_unit(1, "ft", "in", "length"), 12, places=1)

        self.assertAlmostEqual(convert_unit(98765, "m", "km", "length"), 98.765, places=3)
        self.assertAlmostEqual(convert_unit(123456, "km", "mi", "length"), 76712.002, places=2)
        self.assertAlmostEqual(convert_unit(999999, "ft", "mi", "length"), 189.394, places=2)
        self.assertAlmostEqual(convert_unit(54321, "in", "m", "length"), 1379.753, places=2)

    def test_mass_conversions(self):
        from src.mcp_mathematics.calculator import convert_unit

        self.assertAlmostEqual(convert_unit(1, "kg", "g", "mass"), 1000, places=1)
        self.assertAlmostEqual(convert_unit(1, "kg", "lb", "mass"), 2.20462, places=3)
        self.assertAlmostEqual(convert_unit(1, "lb", "oz", "mass"), 16, places=1)

        self.assertAlmostEqual(convert_unit(87654, "kg", "ton", "mass"), 87.654, places=3)
        self.assertAlmostEqual(convert_unit(999999, "g", "lb", "mass"), 2204.62, places=2)
        self.assertAlmostEqual(convert_unit(123456, "oz", "kg", "mass"), 3499.92, places=2)
        self.assertAlmostEqual(convert_unit(54321, "lb", "ton", "mass"), 24.6396, places=3)

    def test_temperature_conversions(self):
        from src.mcp_mathematics.calculator import convert_unit

        self.assertEqual(convert_unit(0, "C", "F", "temperature"), 32)
        self.assertEqual(convert_unit(100, "C", "F", "temperature"), 212)
        self.assertEqual(convert_unit(0, "C", "K", "temperature"), 273.15)

    def test_edge_cases(self):
        from src.mcp_mathematics.calculator import convert_unit

        self.assertEqual(convert_unit(0, "m", "km", "length"), 0)
        self.assertEqual(convert_unit(0, "C", "F", "temperature"), 32)

        self.assertEqual(convert_unit(-10, "m", "km", "length"), -0.01)
        self.assertEqual(convert_unit(-40, "C", "F", "temperature"), -40)

        with self.assertRaises(ValueError):
            convert_unit(1, "invalid_unit", "m", "length")

    def test_complex_nested_expressions(self):
        result = float(execute_mathematical_computation("sin(cos(tan(pi/6)))"))
        expected = math.sin(math.cos(math.tan(math.pi / 6)))
        self.assertAlmostEqual(result, expected, places=10)

        result = float(execute_mathematical_computation("log(exp(2) * exp(3))"))
        self.assertAlmostEqual(result, 5, places=10)

        result = float(execute_mathematical_computation("sqrt(16) + sqrt(9) * sqrt(4)"))
        self.assertEqual(result, 10)

    def test_complex_trigonometric_identities(self):
        result = float(execute_mathematical_computation("sin(pi/3)^2 + cos(pi/3)^2"))
        self.assertAlmostEqual(result, 1, places=10)

        angle = math.pi / 4
        result = float(execute_mathematical_computation(f"tan({angle})"))
        expected = float(execute_mathematical_computation(f"sin({angle})/cos({angle})"))
        self.assertAlmostEqual(result, expected, places=10)

    def test_advanced_logarithms(self):
        result = float(execute_mathematical_computation("log(256)/log(2)"))
        self.assertAlmostEqual(result, 8, places=5)

        result = float(execute_mathematical_computation("log(1000)/log(10)"))
        self.assertAlmostEqual(result, 3, places=5)

        result = float(execute_mathematical_computation("log(exp(5))"))
        self.assertAlmostEqual(result, 5, places=5)

    def test_factorial_and_combinatorics(self):
        result = float(execute_mathematical_computation("factorial(6)"))
        self.assertAlmostEqual(result, 720, places=5)

        result = float(execute_mathematical_computation("factorial(10)/(factorial(10-3))"))
        self.assertAlmostEqual(result, 720, places=5)

        result = float(
            execute_mathematical_computation("factorial(10)/(factorial(3)*factorial(10-3))")
        )
        self.assertAlmostEqual(result, 120, places=5)

    def test_hyperbolic_complex(self):
        result = float(execute_mathematical_computation("sinh(2) + cosh(2)"))
        expected = math.sinh(2) + math.cosh(2)
        self.assertAlmostEqual(result, expected, places=10)

        result = float(execute_mathematical_computation("tanh(1) * tanh(2)"))
        expected = math.tanh(1) * math.tanh(2)
        self.assertAlmostEqual(result, expected, places=10)

    def test_modulo_operations(self):
        result = float(execute_mathematical_computation("17 % 5"))
        self.assertEqual(result, 2)

        result = float(execute_mathematical_computation("100 % 7"))
        self.assertEqual(result, 2)

        result = float(execute_mathematical_computation("25 % 4"))
        self.assertEqual(result, 1)

    def test_floor_ceil_functions(self):
        result = float(execute_mathematical_computation("floor(3.7)"))
        self.assertEqual(result, 3)

        result = float(execute_mathematical_computation("ceil(3.2)"))
        self.assertEqual(result, 4)

        result = float(execute_mathematical_computation("floor(-2.3)"))
        self.assertEqual(result, -3)

    def test_gcd_lcm_calculations(self):
        result = float(execute_mathematical_computation("gcd(48, 18)"))
        self.assertEqual(result, 6)

        result = float(execute_mathematical_computation("lcm(12, 15)"))
        self.assertEqual(result, 60)

        result = float(execute_mathematical_computation("gcd(100, 75)"))
        self.assertEqual(result, 25)

    def test_complex_unit_chains(self):
        from src.mcp_mathematics.calculator import convert_unit

        meters_to_inches = convert_unit(1, "m", "in", "length")
        self.assertAlmostEqual(meters_to_inches, 39.3701, places=2)

        kg_to_oz = convert_unit(1, "kg", "oz", "mass")
        self.assertAlmostEqual(kg_to_oz, 35.274, places=2)

        hours_to_ms = convert_unit(1, "h", "ms", "time")
        self.assertEqual(hours_to_ms, 3600000)

    def test_temperature_round_trips(self):
        from src.mcp_mathematics.calculator import convert_unit

        celsius = 25
        to_fahrenheit = convert_unit(celsius, "C", "F", "temperature")
        back_to_celsius = convert_unit(to_fahrenheit, "F", "C", "temperature")
        self.assertAlmostEqual(back_to_celsius, celsius, places=10)

        kelvin = 300
        to_celsius = convert_unit(kelvin, "K", "C", "temperature")
        back_to_kelvin = convert_unit(to_celsius, "C", "K", "temperature")
        self.assertAlmostEqual(back_to_kelvin, kelvin, places=10)

    def test_area_volume_conversions(self):
        from src.mcp_mathematics.calculator import convert_unit

        sqm_to_sqft = convert_unit(1, "m2", "ft2", "area")
        self.assertAlmostEqual(sqm_to_sqft, 10.7639, places=2)

        liters_to_gallons = convert_unit(1, "L", "gal", "volume")
        self.assertAlmostEqual(liters_to_gallons, 0.264172, places=4)

        ml_to_cup = convert_unit(250, "mL", "cup", "volume")
        self.assertAlmostEqual(ml_to_cup, 1.05669, places=3)

    def test_energy_power_conversions(self):
        from src.mcp_mathematics.calculator import convert_unit

        joules_to_calories = convert_unit(1000, "J", "cal", "energy")
        self.assertAlmostEqual(joules_to_calories, 239.006, places=2)

        watts_to_hp = convert_unit(1000, "W", "hp", "power")
        self.assertAlmostEqual(watts_to_hp, 1.34102, places=3)

        kwh_to_joules = convert_unit(1, "kWh", "J", "energy")
        self.assertEqual(kwh_to_joules, 3600000)

    def test_pressure_force_conversions(self):
        from src.mcp_mathematics.calculator import convert_unit

        pa_to_psi = convert_unit(100000, "Pa", "psi", "pressure")
        self.assertAlmostEqual(pa_to_psi, 14.5038, places=2)

        newtons_to_lbf = convert_unit(100, "N", "lbf", "force")
        self.assertAlmostEqual(newtons_to_lbf, 22.4809, places=2)

        bar_to_atm = convert_unit(1, "bar", "atm", "pressure")
        self.assertAlmostEqual(bar_to_atm, 0.98692, places=3)

    def test_compound_interest_variations(self):
        from src.mcp_mathematics.calculator import calculate_compound_interest

        result = calculate_compound_interest(5000, 8, 10, 4)
        self.assertAlmostEqual(result["amount"], 11040.20, places=2)

        result = calculate_compound_interest(1000, 12, 5, 12)
        self.assertAlmostEqual(result["amount"], 1816.70, places=2)

        result = calculate_compound_interest(10000, 5, 20, 1)
        self.assertAlmostEqual(result["amount"], 26532.98, places=2)

    def test_loan_payment_scenarios(self):
        from src.mcp_mathematics.calculator import calculate_loan_payment

        result = calculate_loan_payment(200000, 4.5, 30)
        self.assertAlmostEqual(result["payment"], 1013.37, places=2)

        result = calculate_loan_payment(30000, 6, 5)
        self.assertAlmostEqual(result["payment"], 579.98, places=2)

        result = calculate_loan_payment(500000, 3.5, 15)
        self.assertAlmostEqual(result["payment"], 3574.41, places=2)

    def test_percentage_operations(self):
        from src.mcp_mathematics.calculator import calculate_percentage, calculate_percentage_change

        change = calculate_percentage_change(100, 150)
        self.assertEqual(change, 50)

        change = calculate_percentage_change(200, 50)
        self.assertEqual(change, -75)

        result = calculate_percentage(100, 25)
        self.assertEqual(result, 25)

        result = calculate_percentage(100, 15)
        self.assertEqual(result, 15)

    def test_tax_tip_calculations(self):
        from src.mcp_mathematics.calculator import calculate_tax, calculate_tip

        tax_result = calculate_tax(100, 8.5)
        self.assertEqual(tax_result["tax_amount"], 8.5)

        total_with_tax = calculate_tax(250, 10, is_inclusive=False)
        self.assertEqual(total_with_tax["total_amount"], 275)

        tip_result = calculate_tip(80, 20)
        self.assertEqual(tip_result["tip_amount"], 16)

        total_with_tip = calculate_tip(100, 18)
        self.assertEqual(total_with_tip["total_amount"], 118)

    def test_discount_markup_calculations(self):
        from src.mcp_mathematics.calculator import calculate_discount, calculate_markup

        discount_result = calculate_discount(100, 25)
        self.assertEqual(discount_result["final_price"], 75)

        discount_result = calculate_discount(80, 10)
        self.assertEqual(discount_result["final_price"], 72)

        markup_result = calculate_markup(50, 40)
        self.assertEqual(markup_result["selling_price"], 70)

        markup_result = calculate_markup(100, 15)
        self.assertEqual(markup_result["selling_price"], 115)

    def test_bill_splitting_scenarios(self):
        from src.mcp_mathematics.calculator import split_bill

        result = split_bill(120, 4)
        self.assertEqual(result["per_person"], 30)

        result = split_bill(200, 5, tip_percent=20)
        self.assertEqual(result["per_person"], 48)

        result = split_bill(150, 3, tip_percent=18)
        self.assertEqual(result["per_person"], 59)

    def test_speed_conversion_chains(self):
        from src.mcp_mathematics.calculator import convert_unit

        ms_to_mph = convert_unit(10, "m/s", "mph", "speed")
        self.assertAlmostEqual(ms_to_mph, 22.3694, places=2)

        kmh_to_knot = convert_unit(100, "km/h", "knot", "speed")
        self.assertAlmostEqual(kmh_to_knot, 53.9957, places=2)

        fps_to_ms = convert_unit(100, "ft/s", "m/s", "speed")
        self.assertAlmostEqual(fps_to_ms, 30.48, places=2)

    def test_data_storage_conversions(self):
        from src.mcp_mathematics.calculator import convert_unit

        gb_to_mb = convert_unit(5, "GB", "MB", "data")
        self.assertEqual(gb_to_mb, 5000)

        tb_to_gb = convert_unit(2, "TB", "GB", "data")
        self.assertEqual(tb_to_gb, 2000)

        mb_to_bits = convert_unit(1, "MB", "bit", "data")
        self.assertEqual(mb_to_bits, 8000000)

    def test_frequency_angle_conversions(self):
        from src.mcp_mathematics.calculator import convert_unit

        hz_to_khz = convert_unit(5000, "Hz", "kHz", "frequency")
        self.assertAlmostEqual(hz_to_khz, 5, places=10)

        rpm_to_hz = convert_unit(3600, "rpm", "Hz", "frequency")
        self.assertAlmostEqual(round(rpm_to_hz), 60)

        deg_to_rad = convert_unit(180, "deg", "rad", "angle")
        self.assertAlmostEqual(deg_to_rad, math.pi, places=5)

        grad_to_deg = convert_unit(100, "grad", "deg", "angle")
        self.assertAlmostEqual(grad_to_deg, 90, places=10)

    def test_fuel_economy_conversions(self):
        from src.mcp_mathematics.calculator import convert_unit

        mpg_to_lper100km = convert_unit(30, "mpg", "L/100km", "fuel_economy")
        self.assertAlmostEqual(mpg_to_lper100km, 7.84049, places=1)

        lper100km_to_mpg = convert_unit(8, "L/100km", "mpg", "fuel_economy")
        self.assertAlmostEqual(lper100km_to_mpg, 29.4018, places=1)

        km_per_l_to_mpg = convert_unit(10, "km/L", "mpg", "fuel_economy")
        self.assertAlmostEqual(km_per_l_to_mpg, 23.52, places=1)

    def test_smart_unit_detection(self):
        from src.mcp_mathematics.calculator import detect_unit_type

        self.assertEqual(detect_unit_type("meter"), "length")
        self.assertEqual(detect_unit_type("kg"), "mass")
        self.assertEqual(detect_unit_type("celsius"), "temperature")
        self.assertEqual(detect_unit_type("joule"), "energy")
        self.assertEqual(detect_unit_type("pascal"), "pressure")

    def test_unit_alias_resolution(self):
        from src.mcp_mathematics.calculator import resolve_unit_alias

        self.assertEqual(resolve_unit_alias("meters"), "m")
        self.assertEqual(resolve_unit_alias("kilograms"), "kg")
        self.assertEqual(resolve_unit_alias("celsius"), "C")
        self.assertEqual(resolve_unit_alias("pounds"), "lb")
        self.assertEqual(resolve_unit_alias("ounces"), "oz")

    def test_compound_unit_parsing(self):
        pass

    def test_conversion_history_tracking(self):
        from src.mcp_mathematics.calculator import convert_with_history

        result = convert_with_history(100, "m", "ft")
        self.assertAlmostEqual(result, 328.084, places=2)

        result = convert_with_history(50, "kg", "lb")
        self.assertAlmostEqual(result, 110.231, places=2)

    def test_scientific_notation_handling(self):
        from src.mcp_mathematics.calculator import format_scientific_notation

        formatted = format_scientific_notation(1234567)
        self.assertEqual(str(formatted), "1234567")

        formatted = format_scientific_notation(0.00001234)
        self.assertEqual(formatted, "1.234e-05")

        formatted = format_scientific_notation(5.67e23)
        self.assertEqual(formatted, "5.6700e+23")

    def test_special_mathematical_constants(self):
        pi_value = float(execute_mathematical_computation("pi"))
        self.assertAlmostEqual(pi_value, math.pi, places=10)

        e_value = float(execute_mathematical_computation("e"))
        self.assertAlmostEqual(e_value, math.e, places=10)

        tau_value = float(execute_mathematical_computation("tau"))
        self.assertAlmostEqual(tau_value, 2 * math.pi, places=10)

        phi_value = float(execute_mathematical_computation("phi"))
        self.assertAlmostEqual(phi_value, (1 + math.sqrt(5)) / 2, places=10)

    def test_complex_financial_scenarios(self):
        from src.mcp_mathematics.calculator import (
            calculate_compound_interest,
            calculate_simple_interest,
        )

        compound = calculate_compound_interest(10000, 6, 5, 12)
        simple = calculate_simple_interest(10000, 6, 5)
        difference = compound["amount"] - (10000 + simple["interest"])
        self.assertGreater(difference, 0)

        monthly_compound = calculate_compound_interest(1000, 12, 1, 12)
        yearly_compound = calculate_compound_interest(1000, 12, 1, 1)
        self.assertGreater(monthly_compound["amount"], yearly_compound["amount"])

    def test_advanced_percentage_calculations(self):
        from src.mcp_mathematics.calculator import calculate_percentage, calculate_percentage_change

        percent = calculate_percentage(200, 25)
        self.assertEqual(percent, 50)

        percent = calculate_percentage(150, 18)
        self.assertEqual(percent, 27)

        increase = calculate_percentage_change(80, 120)
        self.assertEqual(increase, 50)

        decrease = calculate_percentage_change(100, 75)
        self.assertEqual(decrease, -25)

    def test_mixed_operations(self):
        result = float(execute_mathematical_computation("sin(pi/4) * sqrt(2)"))
        self.assertAlmostEqual(result, 1, places=5)

        result = float(execute_mathematical_computation("log(100) / log(10)"))
        self.assertAlmostEqual(result, 2, places=5)

        result = float(execute_mathematical_computation("exp(log(5) + log(3))"))
        self.assertAlmostEqual(result, 15, places=4)

        result = float(execute_mathematical_computation("2**10 - 1000"))
        self.assertAlmostEqual(result, 24, places=5)

    def test_extreme_values(self):
        result = float(execute_mathematical_computation("factorial(20)"))
        self.assertEqual(result, 2432902008176640000)

        result = float(execute_mathematical_computation("log(1e-10)"))
        self.assertAlmostEqual(result, -23.0259, places=3)

        result = float(execute_mathematical_computation("exp(20)"))
        self.assertAlmostEqual(result, 485165195.4, places=1)

    def test_degree_radian_conversions(self):
        result = float(execute_mathematical_computation("degrees(pi)"))
        self.assertEqual(result, 180)

        result = float(execute_mathematical_computation("radians(180)"))
        self.assertAlmostEqual(result, math.pi, places=10)

        result = float(execute_mathematical_computation("degrees(2*pi)"))
        self.assertEqual(result, 360)

        result = float(execute_mathematical_computation("radians(90)"))
        self.assertAlmostEqual(result, math.pi / 2, places=10)


if __name__ == "__main__":
    unittest.main()
